"""Utilities for interfacing with accelerator hardware. Both OpenCL and CUDA
are supported, as well as multiple devices.

The modules :mod:`cuda` and :mod:`opencl` provide the abstraction layer,
but most code will not import these directly (and it might not be possible
to import them). Instead, use :func:`create_some_context` to set up a context
on whatever device is available.

Attributes
----------
have_cuda : boolean
    True if PyCUDA could be imported (does not guarantee any CUDA devices)
have_opencl : boolean
    True if PyOpenCL could be imported (does not guarantee any OpenCL devices)
"""

import re
import os
import sys
import io
import itertools
from collections import OrderedDict

import numpy as np
import mako.lexer
from mako.template import Template
from mako.lookup import TemplateLookup
import pkg_resources

try:
    import pycuda.driver
    from . import cuda
    have_cuda = True
except ImportError:
    have_cuda = False
try:
    from . import opencl
    have_opencl = True
except ImportError:
    have_opencl = False


def divup(x, y):
    """Divide x by y and round the result upwards"""
    return (x + y - 1) // y


def roundup(x, y):
    """Rounds x up to the next multiple of y"""
    return divup(x, y) * y


class LinenoLexer(mako.lexer.Lexer):
    """A wrapper that inserts #line directives into the source code. It
    is used by passing `lexer_cls` to the mako template constructor.
    """
    def __init__(self, *args, **kw):
        super(LinenoLexer, self).__init__(*args, **kw)
        self.preprocessor.insert(0, self._lineno_preproc)

    @classmethod
    def _escape_filename(cls, filename):
        """Escapes a string for the C preprocessor"""
        return '"' + re.sub(r'([\\"])', r'\\\1', filename) + '"'

    def _lineno_preproc(self, source):
        """Insert #line directives between every line in `source` and return
        the result.
        """
        if self.filename is not None:
            escaped_filename = self._escape_filename(self.filename)
        else:
            escaped_filename = self._escape_filename('<string>')
        lines = source.split('\n')
        # If the last line is \n-terminated, it will cause an empty
        # final string in lines
        if len(lines) > 0 and lines[-1] == '':
            lines.pop()
        out = []
        for i, line in enumerate(lines):
            out.append('#line {0} {1}\n'.format(i + 1, escaped_filename))
            out.append(line + '\n')
        return ''.join(out)


def _make_lookup(extra_dirs):
    dirs = extra_dirs + [pkg_resources.resource_filename(__name__, '')]
    return TemplateLookup(dirs, lexer_cls=LinenoLexer, strict_undefined=True)


# Cached for convenience
_lookup = _make_lookup([])


def build(context, name, render_kws=None, extra_dirs=None, extra_flags=None, source=None):
    """Build a source module from a mako template.

    Parameters
    ----------
    context : |Context|
        Context for which to compile the code
    name : str
        Source file name, relative to the katsdpsigproc module
    render_kws : dict, optional
        Keyword arguments to pass to mako
    extra_dirs : list of `str`, optional
        Extra directories to search for source code
    extra_flags : list, optional
        Flags to pass to the compiler
    source : str, optional
        If specified, provides the source, overriding `name`

    Returns
    -------
    :class:`cuda.Program` or :class:`opencl.Program`
        Compiled module
    """
    if render_kws is None:
        render_kws = {}
    if extra_flags is None:
        extra_flags = []
    if extra_dirs is None:
        lookup = _lookup
    else:
        lookup = _make_lookup(extra_dirs)
    if source is not None:
        template = Template(source, lookup=lookup, lexer_cls=LinenoLexer)
    else:
        template = lookup.get_template(name)
    rendered_source = template.render(
        simd_group_size=context.device.simd_group_size,
        **render_kws)
    return context.compile(rendered_source, extra_flags)


def all_devices():
    """Return a list of all discovered devices"""
    devices = []
    if have_cuda:
        pycuda.driver.init()
        devices.extend(cuda.Device.get_devices())
    if have_opencl:
        devices.extend(opencl.Device.get_devices())
    return devices


def create_some_context(interactive=True, device_filter=None):
    """Create a single-device context, selecting a device automatically. This
    is similar to `pyopencl.create_some_context`. A number of environment
    variables can be set to limit the choice to a single device:

     - KATSDPSIGPROC_DEVICE: device number from amongst all devices
     - CUDA_DEVICE: CUDA device number (compatible with PyCUDA)
     - PYOPENCL_CTX: OpenCL platform and optionally device number (compatible with PyOpenCL)

    The first of these that is encountered takes effect. If it does not exist,
    an exception is thrown.

    Parameters
    ----------
    interactive : boolean
        If true, and `sys.stdin.isatty()` is true, and there are multiple
        choices, it will prompt the user. Otherwise, it will choose the first
        available device, favouring CUDA over OpenCL, then GPU over
        accelerators over other OpenCL devices.
    device_filter : callable, optional
        If specified, each device in turn is passed to it, and it must return
        True to keep the device as a candidate or False to reject it.

    Raises
    ------
    RuntimeError
        If no device could be found or the user made an invalid selection
    """

    def key(device):
        if device.is_cuda:
            return 100
        elif device.is_gpu:
            return 50
        elif device.is_accelerator:
            return 40
        else:
            return 30

    def parse_id(envar):
        if envar in os.environ:
            try:
                num = int(os.environ[envar])
                if num >= 0:
                    return num
            except ValueError:
                pass
        return None

    def parse_id_list(envar, max_parts):
        """Split an environment variable of the form a[:b[:c...]] into
        its components. If the variable is not validly formatted, returns
        `None`.
        """
        if envar in os.environ:
            try:
                fields = os.environ[envar].split(':', max_parts)
                out = []
                for value in fields:
                    num = int(value)
                    if num < 0:
                        raise ValueError('')
                    out.append(num)
                if not out:
                    raise ValueError('')
                return out
            except ValueError:
                pass
        return None

    cuda_id = None
    opencl_id = None
    device_id = None

    device_id = parse_id('KATSDPSIGPROC_DEVICE')
    if device_id is None:
        cuda_id = parse_id('CUDA_DEVICE')
        if cuda_id is None:
            # Fields are platform number and device number
            opencl_id = parse_id_list('PYOPENCL_CTX', 2)

    cuda_devices = []
    opencl_devices = []
    if have_cuda:
        pycuda.driver.init()
        cuda_devices = cuda.Device.get_devices()
    if have_opencl:
        opencl_devices = opencl.Device.get_devices_by_platform()
    if device_filter is not None:
        cuda_devices = list(filter(device_filter, cuda_devices))
        opencl_devices = [list(filter(device_filter, platform)) for platform in opencl_devices]

    try:
        if cuda_id is not None:
            devices = [cuda_devices[cuda_id]]
        elif opencl_id is not None:
            if len(opencl_id) == 1:
                devices = opencl_devices[opencl_id[0]]
            else:
                devices = [opencl_devices[opencl_id[0]][opencl_id[1]]]
        else:
            devices = cuda_devices + list(itertools.chain(*opencl_devices))
            if device_id is not None:
                devices = [devices[device_id]]
    except IndexError:
        raise RuntimeError('Out-of-range device selected')

    if not devices:
        raise RuntimeError('No compute devices found')

    if interactive and len(devices) > 1 and sys.stdin.isatty():
        print("Select device:")
        for i, device in enumerate(devices):
            print("    [{0}]: {1} ({2})".format(i, device.name, device.platform_name))
        print()
        choice = input('Enter selection: ')
        try:
            choice = int(choice)
            if choice < 0:
                raise IndexError   # Otherwise Python's negative indexing kicks in
            device = devices[choice]
        except (ValueError, IndexError):
            raise RuntimeError('Invalid device number')
    else:
        devices.sort(key=key, reverse=True)
        device = devices[0]

    return device.make_context()


class HostArray(np.ndarray):
    """A restricted array class that can be used to initialise a
    :class:`DeviceArray`. It uses C ordering and allows padding, which
    is always in units of the dtype. It optionally uses pinned memory
    to allow fast transfer to and from device memory.

    Because of limitations in numpy and PyCUDA (which do not support
    non-contiguous memory very well), it works by taking a slice from
    the origin of contiguous storage, and using that contiguous storage
    in host-device copies. While one can create views, those views
    cannot be used in fast copies because there is no way to know
    whether the view is anchored at the origin.

    See the numpy documentation on subclassing for an explanation of
    why it is written the way it is.

    Parameters
    ----------
    shape : tuple
        Shape for the array
    dtype : numpy dtype
        Data type for the array
    padded_shape : tuple, optional
        Total size of memory allocation (defaults to `shape`)
    context : |Context|, optional
        If specified, the memory will be allocated in a way that allows
        efficient copies to and from this context.
    """
    def __new__(cls, shape, dtype, padded_shape=None, context=None):
        if padded_shape is None:
            padded_shape = shape
        assert len(padded_shape) == len(shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        if context is not None:
            owner = context.allocate_pinned(padded_shape, dtype)
        else:
            owner = np.empty(padded_shape, dtype)
        if shape:
            index = tuple([slice(0, x) for x in shape])
            obj = owner[index].view(HostArray)
        else:
            obj = owner.view(HostArray)
        obj._owner = owner
        obj.padded_shape = padded_shape
        return obj

    @classmethod
    def safe(cls, obj):
        """Determines whether self can be copied to/from a device
        directly.
        """
        try:
            return obj._owner is not None
        except AttributeError:
            return False

    @classmethod
    def padded_view(cls, obj):
        """Retrieves the view of the full memory without padding. Returns
        `None` if `cls.safe(obj)` is `False`."""
        try:
            return obj._owner
        except AttributeError:
            return None

    def __array_finalize__(self, obj):
        if obj is not None:
            # View casting or created from a template: we cannot vouch for it.
            self._owner = None
            if isinstance(obj, HostArray):
                self.padded_shape = obj.padded_shape
            else:
                self.padded_shape = None


class DeviceArray(object):
    """A light-weight array-like wrapper around a device buffer, that
    handles padding better than PyCUDA (which
    has very poor support).

    It only supports C-order arrays where the inner-most dimension is
    contiguous. Transfers are designed to use an :class:`HostArray` of the
    same shape and padding, but fall back to using a copy when
    necessary.

    Parameters
    ----------
    context : |Context|
        Context in which to allocate the memory
    shape : tuple
        Shape for the usable data
    dtype : numpy dtype
        Data type
    padded_shape : tuple, optional
        Shape for memory allocation (defaults to `shape`)
    raw : low-level allocation from `allocate_raw` method of an allocator
        If specified, provides the backing memory
    """

    def __init__(self, context, shape, dtype, padded_shape=None, raw=None):
        if padded_shape is None:
            padded_shape = shape
        assert len(shape) == len(padded_shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        self.shape = shape
        self.dtype = np.dtype(dtype)
        self.padded_shape = padded_shape
        self.context = context
        self.buffer = context.allocate(padded_shape, dtype, raw)

    @property
    def ndim(self):
        """Number of dimensions"""
        return len(self.shape)

    @property
    def strides(self):
        """Strides, as in numpy"""
        ans = [self.dtype.itemsize]
        for i in range(len(self.padded_shape) - 1, 0, -1):
            ans.append(ans[-1] * self.padded_shape[i])
        return tuple(reversed(ans))

    def _copyable(self, ary):
        """Whether `ary` can be copied to/from this array directly"""
        return (HostArray.safe(ary)
                and ary.dtype == self.dtype
                and ary.shape == self.shape
                and ary.padded_shape == self.padded_shape)

    @classmethod
    def _contiguous(cls, ary):
        """Returns a contiguous view of a copyable array, for passing to
        PyCUDA or PyOpenCL functions (which require a contiguous view).
        """
        return HostArray.padded_view(ary)

    def empty_like(self):
        """Return an array-like object that can be efficiently copied."""
        return HostArray(self.shape, self.dtype, self.padded_shape, context=self.context)

    def asarray_like(self, ary):
        """Return an array with the same content as `ary`, but the same memory
        layout as self.
        """
        assert ary.shape == self.shape
        if self._copyable(ary):
            return ary
        tmp = self.empty_like()
        np.copyto(tmp, ary, casting='no')
        return tmp

    def set(self, command_queue, ary):
        """Synchronous copy from `ary` to self"""
        ary = self.asarray_like(ary)
        command_queue.enqueue_write_buffer(self.buffer, self._contiguous(ary))

    def get(self, command_queue, ary=None):
        """Synchronous copy from self to `ary`. If `ary` is None,
        or if it is not suitable as a target, the copy is to a newly
        allocated :class:`HostArray`. The actual target is returned.
        """
        if ary is None or not self._copyable(ary):
            ary = self.empty_like()
        command_queue.enqueue_read_buffer(self.buffer, self._contiguous(ary))
        return ary

    def set_async(self, command_queue, ary):
        """Asynchronous copy from `ary` to self"""
        ary = self.asarray_like(ary)
        command_queue.enqueue_write_buffer(
            self.buffer, self._contiguous(ary), blocking=False)

    def get_async(self, command_queue, ary=None):
        """Asynchronous copy from self to `ary` (see `get`)."""
        if ary is None or not self._copyable(ary):
            ary = self.empty_like()
        command_queue.enqueue_read_buffer(
            self.buffer, self._contiguous(ary), blocking=False)
        return ary

    @classmethod
    def _canonical_slice(cls, region, shape, strides):
        """Transforms a slice selection into a form that is easier to consume
        internally.

        The source and destination region are specified using a limited subset
        of numpy indexing syntax. The following are supported:

        - slices with positive strides
        - integers
        - :code:`np.newaxis`
        - If fewer indices than axes are specified, all elements on the
          remaining axes are used.

        Ellipses are not yet supported, but it would be straightforward to add
        support.

        Parameters
        ----------
        region : int, slice, `np.newaxis` or tuple of same
            Index expression generated by `np.s_` or `np.index_exp`
        shape : sequence
            Shape of the array being indexed
        strides : sequence
            Strides of the array being indexed

        Returns
        -------
        origin : int
            Byte offset for the first selected element
        shape : tuple of int
            Shape of the region to copy
        strides : tuple of int
            Byte stride between elements along each axis of `shape`. Each stride
            is a multiple of the following ones, except that a dimension of
            length 1 may have a stride of 0.
        """
        origin = 0
        self_strides = strides
        self_shape = shape
        shape = []
        strides = []
        axis = 0
        if not isinstance(region, tuple):
            region = (region,)
        for index in region:
            if index is np.newaxis:
                shape.append(1)
                strides.append(0)
            elif isinstance(index, slice):
                if axis >= len(self_shape):
                    raise IndexError('Too many axes in index expression')
                start, stop, stride = index.indices(self_shape[axis])
                if stride <= 0:
                    raise IndexError('Only positive strides are supported')
                origin += start * self_strides[axis]
                shape.append((stop - start) // stride)
                if shape[-1] <= 0:
                    raise IndexError('Empty slice selection')
                strides.append(stride * self_strides[axis])
                axis += 1
            elif isinstance(index, int):
                if axis >= len(self_shape):
                    raise IndexError('Too many axes in index expression')
                if index < 0:
                    index += self_shape[axis]
                if index < 0 or index >= self_shape[axis]:
                    raise IndexError('Index out of range')
                origin += index * self_strides[axis]
                axis += 1
            else:
                raise TypeError('Invalid type in slice: {}'.format(type(index)))
        while axis < len(self_shape):
            shape.append(self_shape[axis])
            strides.append(self_strides[axis])
            axis += 1
        return origin, shape, strides

    @classmethod
    def _region_transfer_params(cls, src, dest, src_region, dest_region):
        """Compute offsets, strides and shape for a copy. Where possible, the
        dimension of the copy is reduced by exploiting contiguity. When
        interpreting the return values, the arrays should be thought of as byte
        arrays, ignoring any shape, padding or dtype of the originals.

        Parameters
        ----------
        src,dest : array-like
            Source and destination buffers, with `dtype, `shape` and `strides`
            attributes
        src_region,dest_region : int, slice, `np.newaxis` or tuple of same
            Index expression generated by `np.s_` or `np.index_exp`

        Returns
        -------
        src_origin,dest_origin : int
            Byte offset into the source/destination array for the start of the copy
        shape : list
            Shape for the copy, which may have fewer dimensions than the arrays.
        src_strides,dest_strides : list
            Byte strides for source/destination.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions are not the same shape
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if src.dtype != dest.dtype:
            raise TypeError('dtypes do not match ({} and {})'.format(src.dtype, dest.dtype))
        src_origin, src_shape, src_strides = cls._canonical_slice(
            src_region, src.shape, src.strides)
        dest_origin, dest_shape, dest_strides = cls._canonical_slice(
            dest_region, dest.shape, dest.strides)
        if src_shape != dest_shape:
            raise ValueError('Source and destination shapes for the copy do not match')
        # Search for axes that can be collapsed together
        new_shape = [src.dtype.itemsize]
        new_src_strides = [1]
        new_dest_strides = [1]
        for axis in range(len(src_shape) - 1, -1, -1):
            if src_shape[axis] == 1:
                continue    # Can just ignore this axis
            if (src_strides[axis] == new_shape[-1] * new_src_strides[-1]
                    and dest_strides[axis] == new_shape[-1] * new_dest_strides[-1]):
                new_shape[-1] *= src_shape[axis]
            else:
                new_shape.append(src_shape[axis])
                new_src_strides.append(src_strides[axis])
                new_dest_strides.append(dest_strides[axis])
        return src_origin, dest_origin, new_shape, new_src_strides, new_dest_strides

    @classmethod
    def _transfer_region(cls, func, buffer1, buffer2, origin1, origin2,
                         shape, strides1, strides2, **kwargs):
        """Wrapper around the command-queue's copy that handles high dimensions
        by looping.
        """
        if len(shape) > 3:
            for i in range(shape[-1]):
                cls._transfer_region(
                    func, buffer1, buffer2,
                    origin1 + strides1[-1] * i,
                    origin2 + strides2[-1] * i,
                    shape[:-1], strides1[:-1], strides2[:-1], **kwargs)
        else:
            func(buffer1, buffer2, origin1, origin2, shape, strides1, strides2, **kwargs)

    def copy_region(self, command_queue, dest, src_region, dest_region):
        """Perform a device-to-device copy of a subregion of `self` to `dest`.
        If the source and destination memory overlap, the result is undefined.

        See :meth:`_canonical_slice` for a description of what slice
        expressions are supported.

        Parameters
        ----------
        command_queue : |CommandQueue|
            Command queue for the asynchronous operation.
        dest : :class:`DeviceArray`
            Target of the copy
        src_region,dest_region : tuple
            Index expressions constructed by `np.s_` or `np.index_exp`.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        src_origin, dest_origin, shape, src_strides, dest_strides = \
            self._region_transfer_params(self, dest, src_region, dest_region)
        self._transfer_region(
            command_queue.enqueue_copy_buffer_rect,
            self.buffer, dest.buffer, src_origin, dest_origin,
            shape, src_strides, dest_strides)

    def get_region(self, command_queue, ary, device_region, ary_region, blocking=True):
        """Perform a device-to-host copy of a subregion of `self` to `ary`.

        See :meth:`_canonical_slice` for a description of what slice
        expressions are supported.

        Parameters
        ----------
        command_queue : |CommandQueue|
            Command queue for the operation.
        ary : :class:`DeviceArray`
            Target of the copy
        device_region,ary_region : tuple
            Index expressions constructed by `np.s_` or `np.index_exp`, to specify
            the source and target regions.
        blocking : bool, optional
            If false, the operation will be asynchronous.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        ValueError
            if `ary` is not a :class:`HostArray` or is a view of one
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if not HostArray.safe(ary):
            raise ValueError('Target region is not suitable for device-to-host copy')
        device_origin, ary_origin, shape, device_strides, ary_strides = \
            self._region_transfer_params(self, ary, device_region, ary_region)
        self._transfer_region(
            command_queue.enqueue_read_buffer_rect,
            self.buffer, self._contiguous(ary), device_origin, ary_origin,
            shape, device_strides, ary_strides, blocking=blocking)

    def set_region(self, command_queue, ary, device_region, ary_region, blocking=True):
        """Perform a host-to-device copy of a subregion `ary` to `self`.

        See :meth:`_canonical_slice` for a description of what slice
        expressions are supported.

        Parameters
        ----------
        command_queue : |CommandQueue|
            Command queue for the operation.
        ary : :class:`DeviceArray`
            Source of the copy
        device_region,ary_region : tuple
            Index expressions constructed by `np.s_` or `np.index_exp`, to specify
            the target and source regions.
        blocking : bool, optional
            If false, the operation will be asynchronous.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if not HostArray.safe(ary):
            tmp = HostArray(ary[ary_region].shape, ary.dtype, context=self.context)
            np.copyto(tmp, ary[ary_region], casting='no')
            ary = tmp
            ary_region = np.s_[()]
        ary_origin, device_origin, shape, ary_strides, device_strides = \
            self._region_transfer_params(ary, self, ary_region, device_region)
        self._transfer_region(
            command_queue.enqueue_write_buffer_rect,
            self.buffer, self._contiguous(ary), device_origin, ary_origin,
            shape, device_strides, ary_strides, blocking=blocking)

    def zero(self, command_queue):
        """Memset with zeros (asynchronously)"""
        command_queue.enqueue_zero_buffer(self.buffer)


class SVMArray(HostArray, DeviceArray):
    """An array that uses shared virtual memory (aka managed memory) to be
    accessible from both the host and the device. It should not be used as
    a source or target for copies to the device, since it already resides on
    the device.

    Due to limitations in PyOpenCL, this is currently only available for CUDA,
    and CUDA's restrictions on managed memory apply. Only the base array (not
    views) can be passed to kernels.

    Parameters
    ----------
    context : |Context|
        Context in which to allocate the memory
    shape : tuple
        Shape for the array
    dtype : numpy dtype
        Data type for the array
    padded_shape : tuple, optional
        Total size of memory allocation (defaults to `shape`)
    raw : PyCUDA DeviceAllocation or PyOpenCL Buffer, optional
        If specified, provides the backing memory
    """
    def __new__(cls, context, shape, dtype, padded_shape=None, raw=None):
        if padded_shape is None:
            padded_shape = shape
        assert len(padded_shape) == len(shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        owner = context.allocate_svm(padded_shape, dtype, raw=raw)
        if shape:
            index = tuple([slice(0, x) for x in shape])
            obj = owner[index].view(SVMArray)
        else:
            obj = owner.view(SVMArray)
        obj._owner = owner
        obj.padded_shape = padded_shape
        obj.context = context
        return obj

    def __init__(self, *args, **kwargs):
        # Suppress DeviceArray's constructor, since we did everything in __new__
        pass

    @property
    def buffer(self):
        return self._owner

    def _copyable(self, ary):
        """Whether `ary` can be copied to/from this array directly"""
        return (ary.dtype == self.dtype
                and ary.shape == self.shape)

    def set(self, command_queue, ary):
        """Synchronous copy from `ary` to self. For SVMArray, this
        is a CPU copy."""
        self[...] = ary

    def get(self, command_queue, ary=None):
        """Synchronous copy from self to `ary`. If `ary` is None,
        or if it is not suitable as a target, the copy is to a newly
        allocated :class:`HostArray`. The actual target is returned.
        For SVMArray, this is a CPU copy.
        """
        if ary is None or not self._copyable(ary):
            return self.copy()
        else:
            ary[:] = self
            return ary

    def set_async(self, command_queue, ary):
        """Asynchronous copy from `ary` to self. This is implemented
        synchronously for SVMArray, but exists for compatibility."""
        self.set(command_queue, ary)

    def get_async(self, command_queue, ary=None):
        """Asynchronous copy from self to `ary` (see `get`). This
        is implemented synchronously for SVMArray, but exists for
        compatibility."""
        return self.get(command_queue, ary)

    def set_region(self, command_queue, ary, device_region, ary_region, blocking=True):
        self[device_region] = ary[ary_region]

    def get_region(self, command_queue, ary, device_region, ary_region, blocking=True):
        ary[ary_region] = self[device_region]


class DeviceAllocator(object):
    """Allocates DeviceArray objects from a context"""
    def __init__(self, context):
        self.context = context

    def allocate(self, shape, dtype, padded_shape=None, raw=None):
        return DeviceArray(self.context, shape, dtype, padded_shape, raw)

    def allocate_raw(self, n_bytes):
        return self.context.allocate_raw(n_bytes)


class SVMAllocator(object):
    """Allocates SVMArray objects from a context"""
    def __init__(self, context):
        self.context = context

    def allocate(self, shape, dtype, padded_shape=None, raw=None):
        return SVMArray(self.context, shape, dtype, padded_shape, raw)

    def allocate_raw(self, n_bytes):
        return self.context.allocate_svm_raw(n_bytes)


class Dimension(object):
    """A single dimension of an :class:`IOSlot`, representing padding and
    alignment requirements. Instances can be linked together, with all
    linked instances (whether directly or indirectly linked) exposing the
    intersection of their requirements. Internally this is represented
    using a union-find tree with path compression.

    Note that `min_padded_round` and `alignment` provide similar but
    different requirements. The former is used to determine a minimum amount
    of padding, but any larger amount is acceptable, even if not a multiple of
    `min_padded_round`. The latter is stricter, always requiring a multiple of
    this value. The latter is not expected to be frequently used except when
    type-punning.

    Dimensions can also be frozen, which is done by :class:`IOSlot` once a
    buffer is bound. This prevents the requirements from changing later (via
    linking or otherwise) and hence invalidating the bound buffer, or
    allowing two buffers with a shared dimension but different strides to
    be used.

    Parameters
    ----------
    size : int
        Size of the actual data, before padding
    min_padded_round : int, optional
        The `min_padded_size` will be computed by rounding `size` up to a
        multiple of this.
    min_padded_size : int, optional
        Minimum size of the padded allocation, overriding `min_padded_round`.
    alignment : int, optional
        Padded size is required to be a multiple of this.
        This must be a power of 2.
    align_dtype : numpy dtype, optional
        If specified, it is a hint that this data type is the fastest-varying
        axis of a multidimensional array. The padded size may be chosen to be
        such that the stride is a multiple of `ALIGN_BYTES`, to ensure
        efficient access by GPUs. The hint will be ignored if `exact` is true.
    exact : bool, optional
        If true, padding is forbidden
    """
    ALIGN_BYTES = 128

    @classmethod
    def _is_power2(cls, value):
        return value > 0 and (value & (value - 1)) == 0

    def __init__(self, size,
                 min_padded_round=None, min_padded_size=None,
                 alignment=1, align_dtype=None, exact=False):
        if min_padded_size is None:
            if min_padded_round is not None:
                min_padded_size = roundup(size, min_padded_round)
            else:
                min_padded_size = size

        if not self._is_power2(alignment):
            raise ValueError('alignment is not a power of 2')
        if min_padded_size < size:
            raise ValueError('padded size is less than size')

        self._parent = None
        self._size = size
        self._min_padded_size = min_padded_size
        self._alignment = alignment
        self._alignment_hint = alignment
        self._exact = exact
        self._frozen = False
        if align_dtype is not None:
            self.add_align_dtype(align_dtype)

    def _root(self):
        if self._parent is None:
            return self
        else:
            self._parent = self._parent._root()
            return self._parent

    @property
    def size(self):
        return self._root()._size

    @property
    def min_padded_size(self):
        return self._root()._min_padded_size

    @property
    def alignment(self):
        return self._root()._alignment

    @property
    def alignment_hint(self):
        return self._root()._alignment_hint

    @property
    def exact(self):
        return self._root()._exact

    @property
    def frozen(self):
        return self._root()._frozen

    def required_padded_size(self):
        """Padded size required to satisfy this dimension"""
        root = self._root()
        size = roundup(root._min_padded_size, root._alignment)
        # If the size is less than the alignment hint, it could
        # waste a huge amount of memory to pad it
        if not root._exact and size >= root._alignment_hint:
            size = roundup(size, root._alignment_hint)
        return size

    def valid(self, padded_size):
        """Whether `size` is a valid padded size"""
        root = self._root()
        if root._exact:
            return padded_size == root.required_padded_size()
        else:
            return (padded_size >= root._min_padded_size
                    and padded_size % root._alignment == 0)

    def add_align_dtype(self, dtype):
        """Add an alignment hint that this will be used with an array whose
        fastest-varying dimension is of type `dtype`. If the size is not
        a power of 2, it is ignored.
        """
        if self.frozen:
            raise ValueError('cannot modify a frozen requirement')
        itemsize = np.dtype(dtype).itemsize
        if self._is_power2(itemsize):
            root = self._root()
            root._alignment_hint = max(root._alignment_hint, self.ALIGN_BYTES // itemsize)

    def link(self, other):
        """Make both `self` and `other` reference a single shared
        requirement.

        Raises
        ------
        ValueError
            If the resulting requirement is exact and unsatisfiable
        """
        root1 = self._root()
        root2 = other._root()
        if root1._frozen or root2._frozen:
            raise ValueError('cannot link frozen requirements')
        if root1 is root2:
            return
        if root1._size != root2._size:
            raise ValueError('sizes are incompatible')
        if root1._exact:
            actual1 = root1.required_padded_size()
            if not root2.valid(actual1):
                raise ValueError('linked requirement is unsatisfiable')
        if root2._exact:
            actual2 = root2.required_padded_size()
            if not root2.valid(actual2):
                raise ValueError('linked requirement is unsatisfiable')
        root1._min_padded_size = max(root1._min_padded_size, root2._min_padded_size)
        root1._alignment = max(root1._alignment, root2._alignment)
        root1._alignment_hint = max(root1._alignment_hint, root2._alignment_hint)
        root1._exact = root1._exact or root2._exact
        root2._parent = root1
        # Clear properties on the child, so that they can't accidentally be used
        # _size can stay as it is immutable
        del root2._min_padded_size
        del root2._alignment
        del root2._alignment_hint
        del root2._exact

    def freeze(self):
        """Prevent further modifications"""
        self._root()._frozen = True


class IOSlotBase(object):
    """An input/output slot of an operation. A slot can be bound to storage,
    or can allocate storage itself. This base class is untyped and unshaped,
    so in most cases one will use :class:`IOSlot` instead.

    Slots are arranged in a tree, and only the root can be manipulated
    directly. The entire tree shares the same storage.
    A slot cannot be reattached to a new parent.
    """
    def __init__(self):
        self.is_root = True

    def check_root(self):
        """Check whether this slot is a root slot, and raise an exception if not."""
        if not self.is_root:
            raise ValueError('not a root slot')

    def required_bytes(self):
        """Number of bytes of device storage required"""
        raise NotImplementedError('abstract base class')

    def is_bound(self):
        """Whether storage is currently attached to this slot"""
        raise NotImplementedError('abstract base class')

    def attachable(self):
        """Whether this slot can be attached as a child to another"""
        return self.is_root and not self.is_bound()

    def _allocate(self, context, raw=None):
        """Variant of :meth:`allocate` that does not check whether this is a
        root slot."""
        raise NotImplementedError('abstract base class')

    def allocate(self, allocator, raw=None):
        """Allocate and immediately bind a buffer satisfying the requirements.

        .. warning:: When `raw` is provided, there is no check that the storage is large enough.

        Parameters
        ----------
        allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`
            Memory allocator from which to obtain the memory
        raw : PyCUDA DeviceAllocation or PyOpenCL Buffer, optional
            Backing storage for the allocation

        Raises
        ------
        ValueError
            If this is not a root slot
        """
        self.check_root()
        return self._allocate(allocator, raw)


class IOSlot(IOSlotBase):
    """An input/output slot with type and shape information. It contains a
    reference to a buffer (initially `None`) and the shape, type, padding and
    alignment requirements for that buffer. It can allocate the buffer on
    request. Each dimension is represesented by a :class:`Dimension`.

    Slots are arranged in a tree, and only the root can be manipulated
    directly. A slot cannot be reattached to a new parent.

    Parameters
    ----------
    dimensions : tuple of int or :class:`Dimension`
        The dimensions of the slots. Integers are converted to
        :class:`Dimension` with no further requirements.
    dtype : numpy dtype
        Type of the data elements
    """

    @classmethod
    def _make_dimension(cls, dimension):
        if isinstance(dimension, Dimension):
            return dimension
        else:
            return Dimension(dimension)

    def __init__(self, dimensions, dtype):
        super(IOSlot, self).__init__()
        self.dimensions = tuple([self._make_dimension(x) for x in dimensions])
        self.shape = tuple([x.size for x in self.dimensions])
        if len(self.dimensions) > 1:
            self.dimensions[-1].add_align_dtype(dtype)
        self.dtype = np.dtype(dtype)
        self.buffer = None

    def is_bound(self):
        return self.buffer is not None

    def validate(self, buffer):
        """Check that `buffer` is suitable for binding.

        Parameters
        ----------
        buffer : :class:`DeviceArray`
            Buffer to validate (must not be `None`)

        Raises
        ------
        TypeError
            If the data type does not match
        ValueError
            If the dimensions or shape do not match
        ValueError
            If a padded size does not match :meth:`Dimension.required_padded_size`
        """
        if buffer.dtype != self.dtype:
            raise TypeError('dtype does not match')
        if len(buffer.shape) != len(self.shape):
            raise ValueError('number of dimensions does not match')
        for size, padded_size, dimension in zip(buffer.shape, buffer.padded_shape, self.dimensions):
            if size != dimension.size:
                raise ValueError('size does not match')
            if padded_size != dimension.required_padded_size():
                raise ValueError('padded size does not match')

    def _bind(self, buffer):
        """Variant of :meth:`bind` that does not check whether this is a root
        slot.
        """
        if buffer is not None:
            self.validate(buffer)
        self.buffer = buffer
        for dimension in self.dimensions:
            dimension.freeze()

    def bind(self, buffer):
        """Set the internal buffer reference. Always use this function rather
        that writing it directly.

        If the buffer is not `None`, it is validated (see :meth:`validate`).

        Parameters
        ----------
        buffer : :class:`DeviceArray` or `None`
            Buffer to store

        Raises
        ------
        ValueError
            If this is not a root slot
        """
        self.check_root()
        self._bind(buffer)

    def required_padded_shape(self):
        """Padded shape required to satisfy only this slot"""
        return tuple([x.required_padded_size() for x in self.dimensions])

    def required_bytes(self):
        return int(np.product(self.required_padded_shape()) * self.dtype.itemsize)

    def _allocate(self, allocator, raw=None):
        buffer = allocator.allocate(self.shape, self.dtype, self.required_padded_shape(), raw=raw)
        self._bind(buffer)
        return buffer


class CompoundIOSlot(IOSlot):
    """IO slot that owns multiple child slots, and presents the combined
    requirement. The children must all have the same type and shape. This
    is used for connecting a single buffer to multiple operations.

    Parameters
    ----------
    children : sequence of :class:`IOSlot`
        Child slots

    Raises
    ------
    ValueError
        If `children` is empty, or if elements have inconsistent shapes
    ValueError
        If any child is not attachable
    TypeError
        If `children` contains inconsistent data types
    """

    def __init__(self, children):
        self.children = list(children)
        if not len(self.children):
            raise ValueError('empty child list')
        shape = children[0].shape
        dtype = children[0].dtype
        dimensions = children[0].dimensions
        # Validate consistency and compute combined requirements
        for child in children:
            if not child.attachable():
                raise ValueError('child is not attachable')
            if child.shape != shape:
                raise ValueError('inconsistent shapes')
            if child.dtype != dtype:
                raise TypeError('inconsistent dtypes')
            for dimension in child.dimensions:
                if dimension.frozen:
                    raise ValueError('child has frozen dimensions')
        for child in children:
            for x, y in zip(dimensions, child.dimensions):
                x.link(y)
        super(CompoundIOSlot, self).__init__(dimensions, dtype)
        for child in children:
            child.is_root = False

    def _bind(self, buffer):
        super(CompoundIOSlot, self)._bind(buffer)
        for child in self.children:
            child._bind(buffer)


class AliasIOSlot(IOSlotBase):
    """Slot that aggregates multiple child slots (which need not have the same
    type or shape), and allocates a single low-level buffer to back all of them.

    This is typically used when logically distinct slots can share memory
    because they do not contain live data at the same time.

    Parameters
    ----------
    children : list of :class:`IOSlotBase`

    Raises
    ------
    ValueError
        If `children` is empty or contains non-attachable elements
    """
    def __init__(self, children):
        super(AliasIOSlot, self).__init__()
        self.children = children
        self.raw = None
        if not len(self.children):
            raise ValueError('empty child list')
        for child in self.children:
            if not child.attachable():
                raise ValueError('child is not attachable')
        for child in self.children:
            child.is_root = False

    def is_bound(self):
        return self.raw is not None

    def required_bytes(self):
        return max([child.required_bytes() for child in self.children])

    def _allocate(self, allocator, raw=None):
        if raw is None:
            raw = allocator.allocate_raw(self.required_bytes())
        for child in self.children:
            child._allocate(allocator, raw)
        self.raw = raw
        return raw


class Operation(object):
    """An instance of a device operation. Typically one first creates a
    template (which contains the program code, and is expensive to create) and
    then instantiates it for use with specific buffers.

    An instance of this class contains *slots*, which are named instances of
    :class:`IOSlot`. The user binds specific buffers to these slots to specify
    the memory used in the operation.

    This class is only useful when subclassed. The subclass will populate
    the slots. Subclasses also provide a `_run` function that handles the
    implementation of `__call__`.

    Operations are arranged in a tree, with internal nodes subclassing
    :class:`OperationSequence`. Internal nodes provide slots that proxy for
    the slots of their children; the children's slots should not be
    manipulated directly. When a parent ceases to exist, its children should
    no longer be used i.e., they cannot be reattached to a new parent.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for the operation
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots

    Attributes
    ----------
    slots : dictionary
        Maps names to slot instances
    hidden_slots : dictionary
        Extra slots which are not root slots, and hence cannot have buffers
        bound to them, but whose buffers can still be referenced. This is
        generally used by :class:`OperationSequence` when several slots are
        aliased to the same memory.
    is_root : boolean
        True if this operation is not part of an :class:`OperationSequence`.
    """
    def __init__(self, command_queue, allocator=None):
        if allocator is None:
            allocator = DeviceAllocator(command_queue.context)
        elif allocator.context is not None:
            if allocator.context is not command_queue.context:
                raise ValueError('command_queue and allocator have different contexts')
        self.slots = {}
        self.hidden_slots = {}
        self.command_queue = command_queue
        self.is_root = True
        self.allocator = allocator

    def bind(self, **kwargs):
        """Bind buffers to slots by keyword. Each keyword argument name
        specifies a slot name.
        """
        for name, buffer in kwargs.items():
            self.slots[name].bind(buffer)

    def ensure_all_bound(self):
        """Make sure that all slots have a buffer bound, allocating if necessary"""
        for slot in self.slots.values():
            if not slot.is_bound():
                slot.allocate(self.allocator)

    def buffer(self, name):
        """Retrieve the buffer bound to a slot. It will consult
        both :attr:`slots` and :attr:`hidden_slots`.

        Parameters
        ----------
        name : str
            Name of the slot to access

        Returns
        -------
        :class:`DeviceArray`
            Buffer bound to slot `name`, or `None` if the slot exists but is unbound

        Raises
        ------
        KeyError
            If no slot with this name exists
        """
        try:
            slot = self.slots[name]
        except KeyError:
            try:
                slot = self.hidden_slots[name]
            except KeyError:
                raise KeyError('no slot named ' + name)
        return slot.buffer

    def required_bytes(self):
        """Number of bytes of device storage required"""
        return sum([x.required_bytes() for x in self.slots.values()])

    def parameters(self):
        """Returns dictionary of configuration options for this operation"""
        return {}

    def _run(self):
        raise NotImplementedError('abstract base class')

    def __call__(self, **kwargs):
        """Run the operation. Any slots that are not already bound will
        allocate a new buffer. Keyword arguments are passed to :meth:`bind`.
        """
        self.bind(**kwargs)
        self.ensure_all_bound()
        return self._run()


class OperationSequence(Operation):
    """Convenience class for setting up an operation that is built up of
    smaller named operations, with mappings of slots to share data.
    Initially, each slot named *slot* in a child named *op* is remapped to a
    parent slot named *op*:*slot*. After this, each set provided in
    `compounds` is removed from the slots and combined into a single compound
    slot. Finally, each set in `aliases` is removed and combined into an
    alias slot.

    For both `compounds` and `aliases`, if a child does not exist, it is
    skipped, and if none of the children exist, no action is taken.

    Parameters
    ----------
    command_queue : |CommandQueue|
        Command queue for the operation
    operations : sequence of 2-tuples
        Name, operation pairs to add. Calling the operation executes them in order
    compounds : mapping of `str` to sequence of `str`, optional
        Names for compound slots, mapped to the original slot names that are replaced
    aliases : mapping of `str` to sequence of `str`, optional
        Names for alias slots, mapped to the original slot names that are replaced
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, command_queue, operations, compounds=None, aliases=None, allocator=None):
        super(OperationSequence, self).__init__(command_queue, allocator)
        self.operations = OrderedDict(operations)
        for (name, operation) in operations:
            if operation.command_queue is not command_queue:
                raise ValueError('child has a different command queue to the parent')
            if not operation.is_root:
                raise ValueError('child already has another parent')
            for (slot_name, slot) in operation.slots.items():
                self.slots[name + ':' + slot_name] = slot
        if compounds is not None:
            for (name, child_names) in compounds.items():
                children = self._extract_slots(child_names, False)
                if children:
                    self.slots[name] = CompoundIOSlot(children)
        if aliases is not None:
            for (name, child_names) in aliases.items():
                children = self._extract_slots(child_names, True)
                if children:
                    self.slots[name] = AliasIOSlot(children)
        for operation in self.operations.values():
            operation.is_root = False

    def _extract_slots(self, names, add_to_hidden):
        """Remove and return the slots with the given names"""
        ans = []
        for name in names:
            try:
                ans.append(self.slots[name])
                del self.slots[name]
                if add_to_hidden:
                    assert name not in self.hidden_slots
                    self.hidden_slots[name] = ans[-1]
            except KeyError:
                # Ignore missing items
                pass
        return ans

    def _run(self):
        for operation in self.operations.values():
            operation()


def _slot_label(slot, name, hide_detail):
    """Produce contents of HTML label for a slot.

    If `hide_detail` is true, it will not show the shapes and dtype. This is used when
    the details will appear on a parent.
    """
    if isinstance(slot, AliasIOSlot):
        return '<b>{name}</b><br/><font color="red">alias slot</font>'.format(name=name)
    elif hide_detail:
        return '<b>{name}</b>'.format(name=name)
    else:
        shape = '×'.join(str(x) for x in slot.shape)
        if not shape:
            shape = 'scalar'
        padded_shape = '×'.join(str(x) for x in slot.required_padded_shape())
        dtype = str(np.dtype(slot.dtype))
        return '<b>{name}</b><br/>{shape}<br/>{padded_shape}<br/>{dtype}'.format(
            name=name, shape=shape, padded_shape=padded_shape, dtype=dtype)


def _visualize_operation(g, operation, path, slot_map, no_detail_slots, edges):
    """Recursive implementation of :func:`visualize_operation`.

    Parameters
    ----------
    g : :class:`graphviz.Graph`
        Graph or subgraph to be updated with the operation. If the operation has
        children, it will be rendered as a subgraph, otherwise as a single slot
        with ports for the slots.
    operation : :class:`Operation`
        Operation to add to the graph
    path : tuple of str
        Hierarchical name of `operation`. The last component is used for visual
        display, and the rest for generating unique node labels.
    slot_map : dict
        Mapping from slots to node names (or port names). It is updated in place
        with all the slots in this operation and its descendants.
    no_detail_slots : set
        Slots that should not show shape and dtype details. It is updated in
        place with any descendant slots of this node.
    edges : set
        Edges to draw, as tuples of two :class:`IOSlotBase`s.
    """

    all_slots = {**operation.slots, **operation.hidden_slots}

    if hasattr(operation, 'operations'):
        # Graphviz treats subgraphs with the prefix 'cluster_' specially. We
        # only want this for real subgraphs, not the root.
        node_name = 'cluster_' + '!'.join(path) if len(path) > 1 else 'root'
        for slot_name, slot in all_slots.items():
            if isinstance(slot, CompoundIOSlot):
                for child in slot.children:
                    no_detail_slots.add(child)
        with g.subgraph(name=node_name) as sub:
            sub.attr('graph', label=path[-1])
            for child_name, child in operation.operations.items():
                _visualize_operation(sub, child, path + (child_name,),
                                     slot_map, no_detail_slots, edges)
            for slot_name, slot in list(all_slots.items()):
                if slot not in slot_map:     # It's already represented in the child
                    # ":" has special meaning to graphviz, so we replace it with "+".
                    full_name = 'slot_{}_{}'.format('!'.join(path), slot_name.replace(':', '+'))
                    slot_map[slot] = full_name
                    label = '<' + _slot_label(slot, slot_name, slot in no_detail_slots) + '>'
                    sub.node(full_name, label=label, shape='box')
                else:
                    del all_slots[slot_name]
    else:
        node_name = 'op_' + '!'.join(path)
        label = io.StringIO()
        label.write('<<table border="1" cellborder="0" rows="*" columns="*">'
                    '<tr><td colspan="{n_slots}">{name}</td></tr><tr>'
                    .format(name=path[-1], n_slots=len(operation.slots)))
        for slot_name, slot in operation.slots.items():
            label.write('<td port="{name}"><font point-size="9"><i>{name}</i></font></td>'
                        .format(name=slot_name))
            slot_map[slot] = node_name + ':' + slot_name.replace(':', '+')
        label.write('</tr></table>>')
        g.node(node_name, label=label.getvalue(), shape='plain')

    for slot in all_slots.values():
        if hasattr(slot, 'children'):
            for child in slot.children:
                edges.add((child, slot))


def visualize_operation(operation, filename):
    """Write a visualization of an :class:`Operation` to file.

    This requires :mod:`graphviz` to be installed.

    Parameters
    ----------
    operation : :class:`Operation`
        Operation or operation sequence to visualize
    filename : str
        Base filename to write. It should end in ``.gv``, and a rendered
        PDF will automatically be produced with a filename ending in
        ``.gv.pdf``.
    """
    from graphviz import Graph

    g = Graph()
    g.attr('graph', clusterrank='local', ranksep='3', compound='yes')
    slot_map = {}
    edges = set()
    _visualize_operation(g, operation, ('root',), slot_map, set(), edges)
    for a, b in edges:
        g.edge(slot_map[a], slot_map[b])
    g.render(filename)



"""Utility to replace variables in a Python function with a fixed value.

This is intended as an aid to using numba, which requires certain things to
be constants and won't accept some things that really are constants (see
https://github.com/numba/numba/issues/2518).
"""

from __future__ import print_function, division, absolute_import
import ast
import inspect
import functools
import textwrap


class ConstantTransformer(ast.NodeTransformer):
    def __init__(self, constants):
        self._constants = constants

    def visit_Name(self, node):
        try:
            value = ast.Num(self._constants[node.id])
            return ast.copy_location(value, node)
        except KeyError:
            return node


def fold(func, **kwargs):
    """For each key in `kwargs`, replace references to that key in the code
    with the corresponding value (which must be a number).

    This is very hacky, and may not work in all cases. Some caveats:
    - The code will be interpreted under the __future__ imports in this
      module, rather than the one where the function is defined
    - The code needs to be available
    - It may go wrong if the function has a decorator

    The names in `kwargs` do not need to resolve to anything when the
    function is defined.
    """
    src = inspect.getsource(func)
    tree = ast.parse(textwrap.dedent(src))
    new_tree = ConstantTransformer(kwargs).visit(tree)
    compiled = compile(new_tree, inspect.getsourcefile(func), 'exec')
    lcls = {}
    exec(compiled, func.func_globals, lcls)
    out = lcls[func.func_name]
    functools.update_wrapper(out, func)
    return out



import sys
import traceback
from nose.tools import assert_equal, assert_raises
import sqlite3
import threading
from unittest import mock
from .. import tune


def test_autotune_basic():
    received = []
    received_lock = threading.Lock()

    def generate(a, b):
        with received_lock:
            received.append((a, b))
        return lambda iters: a * b

    best = tune.autotune(generate, time_limit=0.001, a=[1, 2], b=[7, 3])
    # Autotuning is parallel, so we can't assert anything about the order
    assert_equal([(1, 3), (1, 7), (2, 3), (2, 7)], sorted(received))
    assert_equal({'a': 1, 'b': 3}, best)


def test_autotune_empty():
    with assert_raises(ValueError):
        tune.autotune(lambda x, y: lambda iters: 0, x=[1, 2], y=[])


class CustomError(RuntimeError):
    pass


def test_autotune_some_raise():
    def generate(x):
        if x == 1:
            raise CustomError('x = 1')

        def measure(iters):
            if x == 3:
                raise CustomError('x = 3')
            return -x
        return measure
    best = tune.autotune(generate, x=[0, 1, 2, 3])
    assert_equal({'x': 2}, best)


def generate_raise(x):
    raise CustomError('x = {0}'.format(x))


def test_autotune_all_raise():
    exc_value = None
    exc_info = None
    with assert_raises(CustomError):
        exc_value = None
        try:
            tune.autotune(generate_raise, x=[1, 2, 3])
        except CustomError as e:
            exc_value = e
            exc_info = sys.exc_info()
            raise

    assert_equal('x = 3', str(exc_value))
    # Check that the traceback refers to the original site, not
    # where it was re-raised
    frames = traceback.extract_tb(exc_info[2])
    assert_equal('generate_raise', frames[-1][2])


class TestAutotuner(object):
    """Tests for the `autotuner` decorator. We use mocking to substitute an
    in-memory database, which is made to persist for the test instead of
    being closed each time the decorator is called."""

    def setup(self):
        self.conn = sqlite3.connect(':memory:')
        self.__class__.autotune_mock = mock.Mock()

    def teardown(self):
        self.conn.close()
        del self.__class__.autotune_mock

    @classmethod
    @tune.autotuner(test={'a': 3, 'b': -1})
    def autotune(cls, context, param):
        return cls.autotune_mock(context, param)

    @mock.patch('katsdpsigproc.tune._close_db')
    @mock.patch('katsdpsigproc.tune._open_db')
    def test(self, open_db_mock, close_db_mock):
        open_db_mock.return_value = self.conn
        tuning = {'a': 1, 'b': 2}
        context = mock.NonCallableMock()
        context.device.driver_version = 'mock version'
        context.device.platform_name = 'mock platform'
        context.device.name = 'mock device'
        self.autotune_mock.return_value = tuning
        ret1 = self.autotune(context, 'xyz')
        ret2 = self.autotune(context, 'xyz')
        self.autotune_mock.assert_called_once_with(context, 'xyz')
        assert_equal(tuning, ret1)
        assert_equal(tuning, ret2)



# coding: utf-8
"""Reduction algorithms"""

import numpy as np

from . import accel
from . import tune


class HReduceTemplate(object):
    """Performs reduction along rows in a 2D array. Only commutative reduction
    operators are supported.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : numpy dtype
        Type of data elements
    ctype : str
        Type (in C/CUDA, not numpy) of data elements
    op : str
        C expression to combine two variables, *a* and *b*
    identity : str
        C expression for an identity value for *op*
    extra_code : str, optional
        Arbitrary C code to paste in (for use by *op* or *identity*)

    tuning : dict, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of workitems per workgroup per row
        - wgsy: number of rows to handle per workgroup
    """
    def __init__(self, context, dtype, ctype, op, identity, extra_code='', tuning=None):
        self.context = context
        self.dtype = dtype
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype, op, identity, extra_code)
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.op = op
        self.identity = identity
        self.extra_code = extra_code
        self.program = accel.build(context, 'hreduce.mako', {
            'wgsx': self.wgsx, 'wgsy': self.wgsy, 'type': self.ctype,
            'op': op, 'identity': identity, 'extra_code': extra_code})

    @classmethod
    @tune.autotuner(test={'wgsx': 64, 'wgsy': 4})
    def autotune(cls, context, dtype, ctype, op, identity, extra_code):
        queue = context.create_tuning_command_queue()
        shape = (2048, 1024)
        src = accel.DeviceArray(context, shape, dtype=dtype)
        dest = accel.DeviceArray(context, (shape[0],), dtype=dtype)

        def generate(**kwargs):
            wgs = kwargs['wgsx'] * kwargs['wgsy']
            if wgs < 32 or wgs > 1024:
                raise RuntimeError('Skipping work group size {wgsx}x{wgsy}'.format(**kwargs))
            template = cls(context, dtype, ctype, op, identity, extra_code, kwargs)
            fn = template.instantiate(queue, shape)
            fn.bind(src=src, dest=dest)
            fn.ensure_all_bound()
            return tune.make_measure(queue, fn)

        return tune.autotune(generate,
                             wgsx=[32, 64, 128],
                             wgsy=[1, 2, 4, 8, 16])

    def instantiate(self, *args, **kwargs):
        return HReduce(self, *args, **kwargs)


class HReduce(accel.Operation):
    """
    Concrete instance of :class:`HReduceTemplate`. In each row, the
    elements in the specified column range are reduced using the reduction
    operator supplied to the template.

    .. rubric:: Slots

    **src** : *rows* × *columns*
        Input data
    **dest** : *rows*
        Output reductions

    Parameters
    ----------
    template : :class:`FillTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : 2-tuple of int
        Shape for the source slot
    column_range : 2-tuple of int, optional
        Half-open range of columns to reduce (defaults to the entire array)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, shape, column_range=None, allocator=None):
        if len(shape) != 2:
            raise ValueError('shape must be 2-dimensional')
        if column_range is None:
            column_range = (0, shape[1])
        if column_range[0] < 0 or column_range[1] > shape[1]:
            raise ValueError('column range overflows the array')
        if column_range[0] >= column_range[1]:
            raise ValueError('column range is empty')

        super(HReduce, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('hreduce')
        self.column_range = column_range
        self.slots['src'] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy), shape[1]),
            self.template.dtype)
        self.slots['dest'] = accel.IOSlot(
            (accel.Dimension(shape[0], self.template.wgsy),),
            self.template.dtype)

    def _run(self):
        src = self.buffer('src')
        dest = self.buffer('dest')
        rows_padded = accel.roundup(src.shape[0], self.template.wgsy)
        n_columns = self.column_range[1] - self.column_range[0]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer, dest.buffer,
                np.int32(self.column_range[0]), np.int32(n_columns),
                np.int32(src.padded_shape[1])
            ],
            global_size=(self.template.wgsx, rows_padded),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self):
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['src'].shape,
            'column_range': self.column_range,
            'op': self.template.op,
            'identity': self.template.identity,
            'extra_code': self.template.extra_code
        }



#!/usr/bin/env python
# for nosetest: nosetests katsdpsigproc.test.test_percentile

from __future__ import division, print_function, absolute_import
import time
import numpy as np
from katsdpsigproc import accel
from katsdpsigproc import percentile as perc5

context = accel.create_some_context(True)
queue = context.create_command_queue(profile=True)

data = np.abs(np.random.randn(4000, 5000)).astype(np.float32)

template = perc5.Percentile5Template(context, max_columns=5000)
perc = template.instantiate(queue, data.shape)
perc.ensure_all_bound()
perc.buffer('src').set(queue, data)
start_event = queue.enqueue_marker()
perc()
end_event = queue.enqueue_marker()
out = perc.buffer('dest').get(queue)

t0 = time.time()
expected = np.percentile(data, [0, 100, 25, 75, 50], axis=1, interpolation='lower')
t1 = time.time()
print('gpu:', end_event.time_since(start_event), 'cpu:', t1 - t0)
np.testing.assert_equal(out, expected)



"""Utilities for scheduling device operations with asyncio."""

import asyncio
import logging
import collections


_logger = logging.getLogger(__name__)


async def wait_until(future, when, loop=None):
    """Like :meth:`asyncio.wait_for`, but with an absolute timeout."""
    def ready(*args):
        if not waiter.done():
            waiter.set_result(None)

    if loop is None:
        loop = asyncio.get_event_loop()
    waiter = asyncio.Future(loop=loop)
    timeout_handle = loop.call_at(when, ready)
    # Ensure that the future is really a future, not a coroutine object
    future = asyncio.ensure_future(future, loop=loop)
    future.add_done_callback(ready)
    try:
        await waiter
        if future.done():
            return future.result()
        else:
            future.remove_done_callback(ready)
            future.cancel()
            raise asyncio.TimeoutError()
    finally:
        timeout_handle.cancel()


async def async_wait_for_events(events, loop=None):
    """Coroutine that waits for a list of device events."""
    def wait_for_events(events):
        for event in events:
            event.wait()
    if loop is None:
        loop = asyncio.get_event_loop()
    if events:
        await loop.run_in_executor(None, wait_for_events, events)


class ResourceAllocation(object):
    """A handle representing a future acquisition of a resource. There are two
    ways to make the acquisition current:

     1. Call :meth:`wait`, which returns a future. The result of this future
        is a list of device events that must complete before it is safe to use
        the resource; they can either be waited for on the host (for example,
        using ``run_in_executor``), or by enqueuing a device wait before
        device operations on the resource.
     2. The above steps (with a blocking host wait) can be combined using
        :meth:`wait_events`.

    Instances of this class should never be constructed directly. Instead, use
    :meth:`Resource.acquire`.

    This class implements the context manager protocol, providing the
    underlying resource as the return value. This handles some cleanup if the
    method using the resource raises an exception without releasing the
    resource cleanly.
    """
    def __init__(self, start, end, value, loop):
        self._start = start
        self._end = end
        self._loop = loop
        self.value = value

    def wait(self):
        """Return a future that will be set to a list of device events that
        must be waited for.
        """
        return self._start

    async def wait_events(self):
        """Wait for previous use of the resource to be complete on the host.

        This is a coroutine.
        """
        events = await self._start
        await async_wait_for_events(events, loop=self._loop)

    def ready(self, events=None):
        """Indicate that we are done with the resource, and that subsequent
        acquirers may use it. Note that even if the caller decides that it
        doesn't need to use the resource, it must not call this until the
        resource is ready.

        Parameters
        ----------
        events : list
            Device events that must be waited on before the resource is ready
        """
        if events is None:
            events = []
        self._end.set_result(events)

    def __enter__(self):
        return self.value

    def __exit__(self, exc_type, exc_value, exc_tb):
        if not self._end.done():
            if exc_type is not None:
                self._end.set_exception(exc_value)
            else:
                _logger.warning('Resource allocation was not explicitly made ready')
                self.ready()


class Resource(object):
    """Abstraction of a contended resource, which may exist on a device.

    Passing of ownership is done via futures. Acquiring a resource is a
    non-blocking operation that returns two futures: a future to wait for
    before use, and a future to be signalled with a result when done. The
    value of each of these futures is a (possibly empty) list of device
    events which must be waited on before more device work is scheduled.

    Parameters
    ----------
    value : object
        Underlying resource to manage
    loop : asyncio.BaseEventLoop, optional
        Event loop used for asynchronous operations. If not specified, defaults
        to ``asyncio.get_event_loop()``.

    Attributes
    ----------
    value : object
        Underlying resource passed to the constructor
    """
    def __init__(self, value, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        self._future = asyncio.Future(loop=loop)
        self._future.set_result([])
        self.value = value

    def acquire(self):
        """Indicate intent to acquire the resource. This does not actually
        acquire the resource, but instead returns a handle that can be used to
        acquire and release it later. Acquisitions always occur in the order
        in which calls to :meth:`acquire` are made.

        See :class:`ResourceAcquisition` for further details.
        """
        old = self._future
        self._future = asyncio.Future(loop=self._loop)
        return ResourceAllocation(old, self._future, self.value, loop=self._loop)


class JobQueue(object):
    """Maintains a list of in-flight asynchronous jobs."""
    def __init__(self):
        self._jobs = collections.deque()

    def add(self, job):
        """Append a job to the list. If `job` is a coroutine, it is
        automatically wrapped in a task."""
        self._jobs.append(asyncio.ensure_future(job))

    def clean(self):
        """Remove completed jobs from the front of the queue."""
        while self._jobs and self._jobs[0].done():
            self._jobs.popleft().result()     # Re-throws any exception

    async def finish(self, max_remaining=0):
        """Wait for jobs to finish until there are at most `max_remaining` in
        the queue.

        This is a coroutine.
        """
        while len(self._jobs) > max_remaining:
            await self._jobs.popleft()

    def __len__(self):
        return len(self._jobs)

    def __bool__(self):
        return bool(self._jobs)

    def __contains__(self, item):
        return item in self._jobs


__all__ = ['wait_until', 'async_wait_for_events', 'Resource', 'ResourceAllocation', 'JobQueue']



#!/usr/bin/env python

"""RFI flagging package

The package consists of a number of algorithms for *backgrounding*,
*noise estimation* and *thresholding*. Backgrounding finds a smooth version of
the spectrum, with RFI removed. Noise estimation computes statistics on the
Thresholding applies to the difference between the original spectrum and the
background, and identifies the RFI as deviations that are too large to be
noise.

A backgrounding, a noise estimation and a thresholding algorithm are combined
in either :class:`host.FlaggerHost`, :class:`device.FlaggerDevice`, or
:class:`device.FlaggerHostFromDevice`.
"""

MAD_NORMAL = 1.4826
"""Ratio between `median absolute deviation`_ and standard deviation of a Gaussian distribution.

.. _median absolute deviation: https://en.wikipedia.org/wiki/Median_absolute_deviation
"""



"""Tests for RFI noise estimation algorithms"""

import numpy as np

from .. import host, device
from ...test.test_accel import device_test, force_autotune


_deviations = None
_deviations_big = None
_expected = None


def setup():
    global _deviations, _deviations_big, _expected
    _deviations = np.array(
        [
            [0.0, 3.0, 2.4],
            [1.5, -1.4, 4.6],
            [0.0, 1.1, 3.3],
            [5.0, 0.0, -3.1]
        ]).astype(np.float32)
    _expected = np.array([3.25, 1.4, 3.2]) * 1.4826
    shape = (117, 273)

    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _deviations_big = rs.standard_normal(shape).astype(np.float32)


def test_NoiseEstMADHost():
    noise_est = host.NoiseEstMADHost()
    actual = noise_est(_deviations)
    np.testing.assert_allclose(_expected, actual)


class BaseTestNoiseEstDeviceClass(object):
    @device_test
    def test_result(self, context, queue):
        template = self.factory(context)
        ne_host = template.host_class()
        ne_device = device.NoiseEstHostFromDevice(template, queue)
        noise_host = ne_host(_deviations_big)
        noise_device = ne_device(_deviations_big)
        np.testing.assert_allclose(noise_host, noise_device)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        self.factory(context)


class TestNoiseEstMADDevice(BaseTestNoiseEstDeviceClass):
    def factory(self, context):
        return device.NoiseEstMADDeviceTemplate(context)


class TestNoiseEstMADTDevice(BaseTestNoiseEstDeviceClass):
    def factory(self, context):
        return device.NoiseEstMADTDeviceTemplate(context, 10240)



# coding: utf-8
"""RFI flagging algorithms that run on an accelerator, using OpenCL or CUDA.

The noise estimation and thresholding functions may take data in either
channel-major or baseline-major order (the flags are emitted in the same
order). In the former case, the `transposed` member is `False`, otherwise it is
`True`. The flagger classes automatically detect this and apply a transposition
kernel at the appropriate point.
"""

import enum

import numpy as np

from .. import accel
from .. import tune
from .. import transpose
from ..accel import DeviceArray
from . import host


class BackgroundFlags(enum.Enum):
    NONE = 0
    CHANNEL = 1
    FULL = 2

    def __bool__(self):
        return self != BackgroundFlags.NONE


class BackgroundHostFromDevice(object):
    """Wraps a device-side background template to present the host interface."""
    def __init__(self, template, command_queue):
        self.template = template
        self.command_queue = command_queue

    def __call__(self, vis, flags=None):
        if flags is not None and not self.template.use_flags:
            raise TypeError("flags were provided but not included in the template")
        if flags is None and self.template.use_flags:
            raise TypeError("flags were expected but not provided")
        (channels, baselines) = vis.shape
        fn = self.template.instantiate(self.command_queue, channels, baselines)
        # Trigger allocations
        fn.ensure_all_bound()
        fn.buffer('vis').set(self.command_queue, vis)
        if flags is not None:
            fn.buffer('flags').set(self.command_queue, flags)
        # Do the computation
        fn()
        return fn.buffer('deviations').get(self.command_queue)


class BackgroundMedianFilterDeviceTemplate(object):
    """Device algorithm that applies a median filter to each baseline
    (in amplitude). It is the same algorithm as
    :class:`host.BackgroundMedianFilterHost`, but may give slightly different
    results due to rounding errors when computing complex magnitude.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    width : int
        The kernel width (must be odd)
    is_amplitude : boolean
        If ``True``, the inputs are amplitudes rather than complex visibilities
    use_flags : :class:`BackgroundFlags` or bool
        Specify that flags are taken as input to indicate bad data that should
        not contribute to the backgrounding. The legal values are

        - ``FULL``: takes flags with same shape as visibilities
        - ``CHANNEL``: takes a single flag per channel
        - ``NONE``: do not take flags as input

        For backwards compatibility, ``True`` is an alias for ``CHANNEL`` and
        ``False`` is an alias for ``NONE``.
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: number of work-items per baseline
        - csplit: approximate number of workitems for each channel
    """

    host_class = host.BackgroundMedianFilterHost
    autotune_version = 4

    def __init__(self, context, width, is_amplitude=False,
                 use_flags=BackgroundFlags.NONE, tuning=None):
        if tuning is None:
            tuning = self.autotune(context, width, is_amplitude, use_flags)
        self.context = context
        self.width = width
        self.is_amplitude = is_amplitude
        if use_flags is True:
            use_flags = BackgroundFlags.CHANNEL
        elif use_flags is False:
            use_flags = BackgroundFlags.NONE
        if not isinstance(use_flags, BackgroundFlags):
            raise TypeError('use_flags must be an instance of BackgroundFlags or bool')
        self.use_flags = use_flags
        self.wgs = tuning['wgs']
        self.csplit = tuning['csplit']
        self.program = accel.build(
            context, 'rfi/background_median_filter.mako',
            {
                'width': width, 'wgs': self.wgs,
                'is_amplitude': is_amplitude,
                'BackgroundFlags': BackgroundFlags,
                'use_flags': use_flags
            })

    @classmethod
    @tune.autotuner(test={'wgs': 128, 'csplit': 4})
    def autotune(cls, context, width, is_amplitude, use_flags):
        queue = context.create_tuning_command_queue()
        # Note: baselines must be a multiple of any tested workgroup size
        channels = 4096
        baselines = 8192
        shape = (channels, baselines)
        vis_type = np.float32 if is_amplitude else np.complex64
        vis = DeviceArray(context, shape, vis_type)
        deviations = DeviceArray(context, shape, np.float32)
        if use_flags == BackgroundFlags.CHANNEL:
            flags = DeviceArray(context, (channels,), np.uint8)
        elif use_flags == BackgroundFlags.FULL:
            flags = DeviceArray(context, shape, np.uint8)
        # Initialize with Gaussian random values
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            vis_host = rs.rayleigh(size=shape).astype(np.float32)
        else:
            vis_host = rs.standard_normal(shape) + rs.standard_normal(shape) * 1j
            vis_host = vis_host.astype(np.complex64)
        vis.set(queue, vis_host)
        if use_flags:
            flags.set(queue, (rs.uniform(size=flags.shape) < 0.0001).astype(np.uint8))

        def generate(**tuning):
            fn = cls(context, width, is_amplitude, use_flags, tuning).instantiate(
                queue, channels, baselines)
            fn.bind(vis=vis, deviations=deviations)
            if use_flags:
                fn.bind(flags=flags)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgs=[32, 64, 128, 256, 512], csplit=[1, 2, 4, 8, 16])

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`BackgroundMedianFilterDevice`."""
        return BackgroundMedianFilterDevice(self, *args, **kwargs)


class BackgroundMedianFilterDevice(accel.Operation):
    """Concrete instance of :class:`BackgroundMedianFilterDeviceTemplate`.

    .. rubric:: Slots

    **vis** : channels × baselines, float32 or complex64
        Input visibilities, or their amplitudes if `template.is_amplitude` is true
    **flags** : channels × baselines or channels, float32
        Input flags (only present if `template.use_flags` is used)
    **deviations** : channels × baselines, float32
        Output deviations from the background

    Parameters
    ----------
    template : :class:`BackgroundMedianFilterDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    channels, baselines : int
        Shape of the visibilities array
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, channels, baselines, allocator=None):
        super(BackgroundMedianFilterDevice, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('background_median_filter')
        self.channels = channels
        self.baselines = baselines
        vis_type = np.float32 if template.is_amplitude else np.complex64
        dims = (channels,
                accel.Dimension(baselines, self.template.wgs))
        self.slots['vis'] = accel.IOSlot(dims, vis_type)
        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        if template.use_flags == BackgroundFlags.FULL:
            self.slots['flags'] = accel.IOSlot(dims, np.uint8)
        elif template.use_flags == BackgroundFlags.CHANNEL:
            self.slots['flags'] = accel.IOSlot((channels,), np.uint8)

    def _run(self):
        VT = accel.divup(self.channels, self.template.csplit)
        xblocks = accel.divup(self.baselines, self.template.wgs)
        yblocks = accel.divup(self.channels, VT)

        buffers = [self.buffer('vis'), self.buffer('deviations')]
        if self.template.use_flags:
            buffers.append(self.buffer('flags'))
        stride = buffers[0].padded_shape[1]
        self.command_queue.enqueue_kernel(
            self.kernel,
            [b.buffer for b in buffers] + [
                np.int32(self.channels),
                np.int32(stride),
                np.int32(VT)
            ],
            global_size=(xblocks * self.template.wgs, yblocks),
            local_size=(self.template.wgs, 1))

    def parameters(self):
        return {
            'width': self.template.width,
            'use_flags': self.template.use_flags.name,
            'channels': self.channels,
            'baselines': self.baselines
        }


class NoiseEstHostFromDevice(object):
    """Wraps a device-side noise estimator template to present the host interface"""
    def __init__(self, template, command_queue):
        self.template = template
        self.command_queue = command_queue

    def __call__(self, deviations):
        (channels, baselines) = deviations.shape
        transposed = self.template.transposed
        if transposed:
            deviations = deviations.T
        fn = self.template.instantiate(self.command_queue, channels, baselines)
        # Allocate and populate memory
        fn.ensure_all_bound()
        fn.buffer('deviations').set(self.command_queue, deviations)
        # Perform computation
        fn()
        # Copy back results
        noise = fn.buffer('noise').get(self.command_queue)
        return noise


class NoiseEstMADDeviceTemplate(object):
    """Estimate noise using the median of non-zero absolute deviations.

    In most cases NoiseEstMADTDeviceTemplate is more efficient.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of baselines per workgroup
        - wgsy: number of channels per workgroup
    """

    host_class = host.NoiseEstMADHost
    transposed = False

    def __init__(self, context, tuning=None):
        if tuning is None:
            tuning = self.autotune(context)
        self.context = context
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.program = accel.build(context, 'rfi/madnz.mako',
                                   {'wgsx': self.wgsx, 'wgsy': self.wgsy})

    @classmethod
    @tune.autotuner(test={'wgsx': 32, 'wgsy': 8})
    def autotune(cls, context):
        # TODO: do real autotuning
        return {'wgsx': 32, 'wgsy': 8}

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`NoiseEstMADDevice`."""
        return NoiseEstMADDevice(self, *args, **kwargs)


class NoiseEstMADDevice(accel.Operation):
    """Concrete instantiation of :class:`NoiseEstMADDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : channels × baselines, float32
        Input deviations from the background, computed by a backgrounder
    **noise** : baselines, float32
        Output per-baseline noise estimate

    Parameters
    ----------
    template : :class:`NoiseEstMADDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command-queue in which work will be enqueued
    channels, baselines : int
        Shape of the visibility array
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    transposed = False

    def __init__(self, template, command_queue, channels, baselines, allocator=None):
        super(NoiseEstMADDevice, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('madnz')
        self.channels = channels
        self.baselines = baselines
        baselines_dim = accel.Dimension(baselines, self.template.wgsx)
        self.slots['noise'] = accel.IOSlot((baselines_dim,), np.float32)
        self.slots['deviations'] = accel.IOSlot((channels, baselines_dim), np.float32)

    def _run(self):
        blocks = accel.divup(self.baselines, self.template.wgsx)
        vt = accel.divup(self.channels, self.template.wgsy)
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1]),
                np.int32(vt)
            ],
            global_size=(blocks * self.template.wgsx, self.template.wgsy),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self):
        return {
            'channels': self.channels,
            'baselines': self.baselines
        }


class NoiseEstMADTDeviceTemplate(object):
    """Device-side noise estimation by median of absolute deviations. It
    should give the same results as :class:`NoiseEstMADHost`, up to
    floating-point accuracy. It uses transposed (baseline-major) memory
    order, which allows an entire baseline to be efficiently loaded into
    registers.

    .. note:: There is a tradeoff in selecting the workgroup size: a large
        value gives more parallelism and reduces the register pressure, but
        increases the overhead of reduction operations.

    .. note:: This class may fail for very large numbers of channels (10k can
        definitely be supported), in which case :class:`NoiseEstMADDevice` may be
        used.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    max_channels : int
        Maximum number of channels. Choosing too large a value will
        reduce performance.
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of work-items per baseline
    """

    host_class = host.NoiseEstMADHost
    transposed = True

    def __init__(self, context, max_channels, tuning=None):
        self.context = context
        self.max_channels = max_channels
        if tuning is None:
            tuning = self.autotune(context, max_channels)
        self.wgsx = tuning['wgsx']
        vt = accel.divup(max_channels, self.wgsx)
        self.program = accel.build(context, 'rfi/madnz_t.mako', {'vt': vt, 'wgsx': self.wgsx})

    @classmethod
    @tune.autotuner(test={'wgsx': 128})
    def autotune(cls, context, max_channels):
        queue = context.create_tuning_command_queue()
        baselines = 128
        rs = np.random.RandomState(seed=1)
        host_deviations = rs.uniform(size=(baselines, max_channels)).astype(np.float32)

        def generate(**tuning):
            # Very large values of VT cause the AMD compiler to choke and segfault
            if max_channels > 256 * tuning['wgsx']:
                raise ValueError('wgsx is too small')
            fn = cls(context, max_channels, tuning).instantiate(queue, max_channels, baselines)
            fn.slots['noise'].allocate(fn.allocator)
            deviations = fn.slots['deviations'].allocate(fn.allocator)
            deviations.set(queue, host_deviations)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgsx=[32, 64, 128, 256, 512, 1024])

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`NoiseEstMADTDevice`."""
        return NoiseEstMADTDevice(self, *args, **kwargs)


class NoiseEstMADTDevice(accel.Operation):
    """Concrete instance of :class:`NoiseEstMADTDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : baselines × channels, float32
        Input deviations from the background, computed by a backgrounder
    **noise** : baselines, float32
        Output per-baseline noise estimate

    Parameters
    ----------
    template : :class:`NoiseEstMADTDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command-queue in which work will be enqueued
    channels, baselines : int
        Shape of the visibility array
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    transposed = True

    def __init__(self, template, command_queue, channels, baselines, allocator=None):
        super(NoiseEstMADTDevice, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('madnz_t')
        if channels > self.template.max_channels:
            raise ValueError('channels exceeds max_channels')
        self.channels = channels
        self.baselines = baselines
        self.slots['noise'] = accel.IOSlot((baselines,), np.float32)
        self.slots['deviations'] = accel.IOSlot((baselines, channels), np.float32)

    def _run(self):
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1])
            ],
            global_size=(self.template.wgsx, self.baselines),
            local_size=(self.template.wgsx, 1))

    def parameters(self):
        return {
            'max_channels': self.template.max_channels,
            'baselines': self.baselines,
            'channels': self.channels
        }


class ThresholdHostFromDevice(object):
    """Wraps a device-side thresholder template to present the host interface"""
    def __init__(self, template, command_queue, *args, **kwargs):
        self.template = template
        self.command_queue = command_queue
        self.args = args
        self.kwargs = kwargs

    def __call__(self, deviations, noise):
        (channels, baselines) = deviations.shape
        transposed = self.template.transposed
        if transposed:
            deviations = deviations.T

        fn = self.template.instantiate(self.command_queue, channels, baselines,
                                       *self.args, **self.kwargs)
        # Allocate memory and copy data
        fn.ensure_all_bound()
        fn.buffer('deviations').set(self.command_queue, deviations)
        fn.buffer('noise').set(self.command_queue, noise)
        # Do computation
        fn()
        # Copy back results
        flags = fn.buffer('flags').get(self.command_queue)
        if transposed:
            flags = flags.T
        return flags


class ThresholdSimpleDeviceTemplate(object):
    """Device-side thresholding, operating independently on each sample. It
    should give the same results as :class:`ThresholdSimpleHost`, up to
    floating-point accuracy.

    This class can operate on either transposed or non-transposed inputs,
    depending on a constructor argument.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    transposed : boolean
        Whether inputs and outputs are transposed
    flag_value : int
        Number stored in returned value to indicate RFI
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgsx: number of baselines (channels if `transposed`) per workgroup
        - wgsy: number of channels (baselines if `transposed`) per workgroup
    """

    host_class = host.ThresholdSimpleHost

    def __init__(self, context, transposed, flag_value=1, tuning=None):
        if tuning is None:
            tuning = self.autotune(context)
        self.context = context
        self.transposed = transposed
        self.wgsx = tuning['wgsx']
        self.wgsy = tuning['wgsy']
        self.flag_value = flag_value
        if transposed:
            source_name = 'rfi/threshold_simple_t.mako'
        else:
            source_name = 'rfi/threshold_simple.mako'
        self.program = accel.build(
            context, source_name,
            {'wgsx': self.wgsx, 'wgsy': self.wgsy,
             'flag_value': flag_value})

    @classmethod
    @tune.autotuner(test={'wgsx': 32, 'wgsy': 4})
    def autotune(cls, context):
        # TODO: do real autotuning
        return {'wgsx': 32, 'wgsy': 4}

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`ThresholdSimpleDevice`."""
        return ThresholdSimpleDevice(self, *args, **kwargs)


class ThresholdSimpleDevice(accel.Operation):
    """Concrete instance of :class:`ThresholdSimpleDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : channels × baselines (or transposed), float32
        Input deviations from the background
    **noise** : baselines, float32
        Noise estimates per baseline
    **flags** : channels × baselines (or transposed), uint8
        Output flags

    Parameters
    ----------
    template : :class:`ThresholdSimpleDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command-queue in which work will be enqueued
    n_sigma : float
        Number of (estimated) standard deviations for the threshold
    channels, baselines : int
        Shape of the visibility array
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, template, command_queue, channels, baselines, n_sigma, allocator=None):
        super(ThresholdSimpleDevice, self).__init__(command_queue, allocator)
        self.template = template
        if template.transposed:
            kernel_name = 'threshold_simple_t'
        else:
            kernel_name = 'threshold_simple'
        self.kernel = template.program.get_kernel(kernel_name)
        self.n_sigma = n_sigma
        self.channels = channels
        self.baselines = baselines
        self.transposed = template.transposed

        shape = (baselines, channels) if self.transposed else (channels, baselines)
        dims = (accel.Dimension(shape[0], self.template.wgsy),
                accel.Dimension(shape[1], self.template.wgsx))
        noise_dim = dims[0] if self.transposed else dims[1]

        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        self.slots['noise'] = accel.IOSlot((noise_dim,), np.float32)
        self.slots['flags'] = accel.IOSlot(dims, np.uint8)

    def _run(self):
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        flags = self.buffer('flags')
        stride = deviations.padded_shape[1]

        global_x = accel.roundup(deviations.shape[1], self.template.wgsx)
        global_y = accel.roundup(deviations.shape[0], self.template.wgsy)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                deviations.buffer, noise.buffer, flags.buffer,
                np.int32(stride),
                np.float32(self.n_sigma)
            ],
            global_size=(global_x, global_y),
            local_size=(self.template.wgsx, self.template.wgsy))

    def parameters(self):
        return {
            'n_sigma': self.n_sigma,
            'flag_value': self.template.flag_value,
            'transposed': self.transposed,
            'channels': self.channels,
            'baselines': self.baselines
        }


class ThresholdSumDeviceTemplate(object):
    """A device version of :class:`katsdpsigproc.rfi.host.ThresholdSumHost`.
    It uses transposed data. Performance will be best with a large work
    group size, because of the stencil-like nature of the computation.

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    n_windows : int
        Number of window sizes to use
    flag_value : int
        Number stored in returned value to indicate RFI
    tuning : mapping, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: Number of work items to use per work group
        - vt: Number of elements to process in each work item
    """

    host_class = host.ThresholdSumHost
    transposed = True

    def __init__(self, context, n_windows=4, flag_value=1, tuning=None):
        if tuning is None:
            tuning = self.autotune(context, n_windows)
        wgs = tuning['wgs']
        vt = tuning['vt']
        edge_size = 2 ** n_windows - n_windows - 1
        self.chunk = wgs * vt - 2 * edge_size
        assert self.chunk > 0
        self.context = context
        self.n_windows = n_windows
        self.wgs = wgs
        self.vt = vt
        self.flag_value = flag_value
        self.program = accel.build(
            context, 'rfi/threshold_sum.mako',
            {'wgs': self.wgs,
             'vt': self.vt,
             'windows' : self.n_windows,
             'flag_value': self.flag_value})

    @classmethod
    @tune.autotuner(test={'wgs': 128, 'vt': 3})
    def autotune(cls, context, n_windows):
        queue = context.create_tuning_command_queue()
        channels = 4096
        baselines = 128
        shape = (baselines, channels)
        rs = np.random.RandomState(seed=1)
        deviations = DeviceArray(context, shape, dtype=np.float32)
        deviations.set(queue, rs.uniform(size=deviations.shape).astype(np.float32))
        noise = DeviceArray(context, (baselines,), dtype=np.float32)
        noise.set(queue, rs.uniform(high=0.1, size=noise.shape).astype(np.float32))
        flags = DeviceArray(context, shape, dtype=np.uint8)

        def generate(**tuning):
            template = cls(context, n_windows=n_windows, tuning=tuning)
            fn = template.instantiate(queue, channels, baselines, 11.0)
            fn.bind(deviations=deviations, noise=noise, flags=flags)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate,
                             wgs=[32, 64, 128, 256, 512],
                             vt=[1, 2, 3, 4, 8, 16])

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`ThresholdSumDevice`."""
        return ThresholdSumDevice(self, *args, **kwargs)


class ThresholdSumDevice(accel.Operation):
    """Concrete instance of :class:`ThresholdSumDeviceTemplate`.

    .. rubric:: Slots

    **deviations** : baselines × channels, float32
        Input deviations from the background
    **noise** : baselines, float32
        Noise estimates per baseline
    **flags** : baselines × channels, uint8
        Output flags

    Parameters
    ----------
    template : :class:`ThresholSumDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command-queue in which work will be enqueued
    channels, baselines : int
        Shape of the visibility array
    n_sigma : float
        Number of (estimated) standard deviations for the threshold
    threshold_falloff : float, optional
        Controls rate at which thresholds decrease (ρ in Offringa 2010)
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """
    host_class = host.ThresholdSumHost
    transposed = True

    def __init__(self, template, command_queue, channels, baselines, n_sigma, threshold_falloff=1.2,
                 allocator=None):
        super(ThresholdSumDevice, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel('threshold_sum')
        self.channels = channels
        self.baselines = baselines
        self.n_sigma = [np.float32(n_sigma * pow(threshold_falloff, -i))
                        for i in range(template.n_windows)]
        # For channels we must construct the Dimension here rather than in
        # the IOSlot constructor, so that the deviations and flags share
        # the same object and hence have the same stride.
        dims = (baselines, accel.Dimension(channels))
        self.slots['deviations'] = accel.IOSlot(dims, np.float32)
        self.slots['noise'] = accel.IOSlot((baselines,), np.float32)
        self.slots['flags'] = accel.IOSlot(dims, np.uint8)

    def _run(self):
        deviations = self.buffer('deviations')
        noise = self.buffer('noise')
        flags = self.buffer('flags')

        blocks = accel.divup(self.channels, self.template.chunk)
        args = [deviations.buffer, noise.buffer, flags.buffer,
                np.int32(self.channels), np.int32(deviations.padded_shape[1])]
        args.extend(self.n_sigma)
        self.command_queue.enqueue_kernel(
            self.kernel, args,
            global_size=(blocks * self.template.wgs, self.baselines),
            local_size=(self.template.wgs, 1))

    def parameters(self):
        return {
            'n_sigma': self.n_sigma,
            'flag_value': self.template.flag_value,
            'channels': self.channels,
            'baselines': self.baselines
        }


class FlaggerDeviceTemplate(object):
    """Combine device backgrounder, noise estimation and thresholder
    implementations to create a flagger. The thresholder and/or noise
    estimation may take transposed input, in which case this object will manage
    temporary buffers and the transposition automatically.

    Parameters
    ----------
    background, noise_est, threshold : template types
        The templates for the individual steps
    """
    def __init__(self, background, noise_est, threshold):
        self.background = background
        self.noise_est = noise_est
        self.threshold = threshold
        context = self.background.context
        # Check that all operations work in the same context
        assert self.noise_est.context is context
        assert self.threshold.context is context

        # Create transposition operations if needed
        if noise_est.transposed or threshold.transposed:
            self.transpose_deviations = transpose.TransposeTemplate(context, np.float32, 'float')
        else:
            self.transpose_deviations = None
        if threshold.transposed:
            self.transpose_flags = transpose.TransposeTemplate(context, np.uint8, 'unsigned char')
        else:
            self.transpose_flags = None

    def instantiate(self, *args, **kwargs):
        """Create an instance. See :class:`FlaggerDevice`."""
        return FlaggerDevice(self, *args, **kwargs)


class FlaggerDevice(accel.OperationSequence):
    """Concrete instance of :class:`FlaggerDeviceTemplate`.

    .. rubric:: Slots

    **vis** : channels × baselines, float32 or complex64
        Input visibilities (or amplitudes, if the backgrounder takes amplitudes)
    **noise** : baselines, float32
        Estimate of per-baseline noise
    **flags** : channels × baselines, uint8
        Output flags
    **input_flags** : channels × baseline or channels, uint8
        Input flags. These are only used for estimating the
        background, and are *not* automatically copied into the output. Any
        visibilities marked as flagged here will *not* be flagged as RFI.

        This slot is only present if the backgrounder template has
        `use_flags` set, and the value determines the shape.

    .. rubric:: Temporary slots

    Temporary buffers are presented as slots, which allows them to either
    be set by the user or allocated automatically on first use.

    **deviations** : channels × baselines, float32
        Deviations from the background
    **deviations_t** : baselines × channels, float32, optional
        Transpose of `deviations`
    **flags_t** : baselines × channels, uint8, optional
        Transpose of `flags`

    Parameters
    ----------
    template : :class:`FlaggerDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    channels, baselines : int
        Shape of the visibilities array
    background_args : dict, optional
        Extra keyword arguments to pass to the background instantiation
    noise_est_args : dict, optional
        Extra keyword arguments to pass to the noise estimation instantiation
    threshold_args : dict, optional
        Extra keyword arguments to pass to the threshold instantiation
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`
        Allocator used to allocate unbound slots
    """
    def __init__(self, template, command_queue, channels, baselines,
                 background_args={}, noise_est_args={}, threshold_args={},
                 allocator=None):
        self.template = template
        self.channels = channels
        self.baselines = baselines
        self.background = self.template.background.instantiate(
            command_queue, channels, baselines, allocator=allocator, **background_args)
        self.noise_est = self.template.noise_est.instantiate(
            command_queue, channels, baselines, allocator=allocator, **noise_est_args)
        self.threshold = self.template.threshold.instantiate(
            command_queue, channels, baselines, allocator=allocator, **threshold_args)

        noise_est_suffix = '_t' if self.noise_est.transposed else ''
        threshold_suffix = '_t' if self.threshold.transposed else ''

        operations = []
        compounds = {
            'vis': ['background:vis'],
            'input_flags': ['background:flags'],
            'deviations': ['background:deviations', 'transpose_deviations:src'],
            'deviations_t': ['transpose_deviations:dest'],
            'noise': ['noise_est:noise', 'threshold:noise'],
            'flags_t': ['transpose_flags:src'],
            'flags': ['transpose_flags:dest'],
        }
        compounds['deviations' + noise_est_suffix].append('noise_est:deviations')
        compounds['deviations' + threshold_suffix].append('threshold:deviations')
        compounds['flags' + threshold_suffix].append('threshold:flags')

        operations.append(('background', self.background))
        if self.template.transpose_deviations:
            self.transpose_deviations = self.template.transpose_deviations.instantiate(
                command_queue, (channels, baselines))
            operations.append(('transpose_deviations', self.transpose_deviations))
        operations.append(('noise_est', self.noise_est))
        operations.append(('threshold', self.threshold))
        if self.template.transpose_flags:
            self.transpose_flags = self.template.transpose_flags.instantiate(
                command_queue, (baselines, channels))
            operations.append(('transpose_flags', self.transpose_flags))

        super(FlaggerDevice, self).__init__(
            command_queue, operations, compounds, allocator=allocator)


class FlaggerHostFromDevice(object):
    """Wrapper that makes a :class:`FlaggerDeviceTemplate` present the
    interface of :class:`FlaggerHost`. This is intended only for ease of
    use. It is not efficient, because it allocates and frees memory on
    every call.

    Parameters
    ----------
    template : :class:`FlaggerDeviceTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    background_args : dict, optional
        Extra keyword arguments to pass to the background instantiation
    noise_est_args : dict, optional
        Extra keyword arguments to pass to the noise estimation instantiation
    threshold_args : dict, optional
        Extra keyword arguments to pass to the threshold instantiation
    """
    def __init__(self, template, command_queue,
                 background_args={}, noise_est_args={}, threshold_args={}):
        self.template = template
        self.command_queue = command_queue
        self.background_args = dict(background_args)
        self.noise_est_args = dict(noise_est_args)
        self.threshold_args = dict(threshold_args)

    def __call__(self, vis, input_flags=None):
        if input_flags is not None and not self.template.background.use_flags:
            raise TypeError("channel flags were provided but not included in the template")
        if input_flags is None and self.template.background.use_flags:
            raise TypeError("channel flags were expected but not provided")
        (channels, baselines) = vis.shape
        fn = self.template.instantiate(
            self.command_queue, channels, baselines,
            self.background_args, self.noise_est_args, self.threshold_args)
        fn.ensure_all_bound()
        fn.buffer('vis').set(self.command_queue, vis)
        if input_flags is not None:
            fn.buffer('input_flags').set(self.command_queue, input_flags)
        fn()
        return fn.buffer('flags').get(self.command_queue)



import numpy as np

from .. import host
from ...test.test_accel import device_test, force_autotune
from .. import device


def setup():
    global _vis, _vis_big, _flags, _flags_big
    shape = (417, 313)
    _vis = np.array([[1.25, 1.5j, 1.0, 2.0, -1.75, 2.0]]).T.astype(np.complex64)
    _flags = np.array([0, 0, 1, 0, 0, 4]).T.astype(np.uint8)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _vis_big = (rs.standard_normal(shape) + rs.standard_normal(shape) * 1j).astype(np.complex64)
    _flags_big = (rs.random_sample(shape) < 0.1).astype(np.uint8)
    # Ensure that in some cases the entire window is flagged. Also test with non-0/1
    # flag values.
    _flags_big[100:110, 0:100] = 4


class TestBackgroundMedianFilterHost(object):
    def setup(self):
        self.background = host.BackgroundMedianFilterHost(3)

    def test(self):
        out = self.background(_vis)
        ref = np.array([[-0.125, 0.25, -0.5, 0.25, -0.25, 0.125]]).T.astype(np.float32)
        np.testing.assert_equal(ref, out)

    def test_flags(self):
        out = self.background(_vis, _flags)
        ref = np.array([[-0.125, 0.125, 0.0, 0.125, -0.125, 0.0]]).T.astype(np.float32)
        np.testing.assert_equal(ref, out)


class BaseTestBackgroundDeviceClass(object):
    @device_test
    def test_result(self, context, queue):
        width = 5
        bg_device_template = self.factory(context, width)
        bg_host = bg_device_template.host_class(width, self.amplitudes)
        bg_device = device.BackgroundHostFromDevice(bg_device_template, queue)
        if self.amplitudes:
            vis = np.abs(_vis_big)
        else:
            vis = _vis_big
        if self.use_flags:
            full = self.use_flags == device.BackgroundFlags.FULL
            flags = _flags_big if full else _flags_big[:, 0]
            out_host = bg_host(vis, flags)
            out_device = bg_device(vis, flags)
        else:
            out_host = bg_host(vis)
            out_device = bg_device(vis)
        # Uses an abs tolerance because backgrounding subtracts nearby values
        np.testing.assert_allclose(out_host, out_device, atol=1e-6)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        self.factory(context, 5)


class TestBackgroundMedianFilterDevice(BaseTestBackgroundDeviceClass):
    amplitudes = False
    use_flags = device.BackgroundFlags.NONE

    def factory(self, context, width):
        return device.BackgroundMedianFilterDeviceTemplate(
            context, width, self.amplitudes, self.use_flags)


class TestBackgroundMedianFilterDeviceAmplitudes(TestBackgroundMedianFilterDevice):
    amplitudes = True


class TestBackgroundMedianFilterDeviceFlags(TestBackgroundMedianFilterDevice):
    use_flags = device.BackgroundFlags.CHANNEL



"""On-device transposition of 2D arrays"""

from __future__ import division, print_function, absolute_import
import numpy as np
from . import accel
from . import tune


class TransposeTemplate(object):
    """Kernel for transposing a 2D array of data

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : numpy dtype
        Type of data elements
    ctype : str
        Type (in C/CUDA, not numpy) of data elements
    tuning : dict, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - block: number of workitems per workgroup in each dimension
        - vtx, vty: elements per workitem in each dimension
    """

    autotune_version = 1

    def __init__(self, context, dtype, ctype, tuning=None):
        self.context = context
        self.dtype = np.dtype(dtype)
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype)
        self._block = tuning['block']
        self._tilex = tuning['block'] * tuning['vtx']
        self._tiley = tuning['block'] * tuning['vty']
        self.program = accel.build(context, "transpose.mako", {
            'block': tuning['block'],
            'vtx': tuning['vtx'],
            'vty': tuning['vty'],
            'ctype': ctype
        })

    @classmethod
    @tune.autotuner(test={'block': 8, 'vtx': 2, 'vty': 3})
    def autotune(cls, context, dtype, ctype):
        queue = context.create_tuning_command_queue()
        in_shape = (2048, 2048)
        out_shape = (2048, 2048)
        in_data = accel.DeviceArray(context, in_shape, dtype=dtype)
        out_data = accel.DeviceArray(context, out_shape, dtype=dtype)

        def generate(block, vtx, vty):
            local_mem = (block * vtx + 1) * (block * vty) * np.dtype(dtype).itemsize
            if local_mem > 32768:
                # Skip configurations using lots of lmem
                raise RuntimeError('too much local memory')
            fn = cls(context, dtype, ctype, {
                'block': block,
                'vtx': vtx,
                'vty': vty}).instantiate(queue, in_shape)
            fn.bind(src=in_data, dest=out_data)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate,
                             block=[4, 8, 16, 32],
                             vtx=[1, 2, 3, 4],
                             vty=[1, 2, 3, 4])

    def instantiate(self, *args, **kwargs):
        return Transpose(self, *args, **kwargs)


class Transpose(accel.Operation):
    """Concrete instance of :class:`TransposeTemplate`.

    .. rubric:: Slots

    **src**
        Input

    **dest**
        Output
    """
    def __init__(self, template, command_queue, shape, allocator=None):
        super(Transpose, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("transpose")
        self.shape = shape
        self.slots['src'] = accel.IOSlot(shape, template.dtype)
        self.slots['dest'] = accel.IOSlot((shape[1], shape[0]), template.dtype)

    def _run(self):
        src = self.buffer('src')
        dest = self.buffer('dest')
        # Round up to number of blocks in each dimension
        in_row_tiles = accel.divup(src.shape[0], self.template._tiley)
        in_col_tiles = accel.divup(src.shape[1], self.template._tilex)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                dest.buffer, src.buffer,
                np.int32(src.shape[0]), np.int32(src.shape[1]),
                np.int32(dest.padded_shape[1]),
                np.int32(src.padded_shape[1])
            ],
            global_size=(in_col_tiles * self.template._block, in_row_tiles * self.template._block),
            local_size=(self.template._block, self.template._block))

    def parameters(self):
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['src'].shape
        }



import numpy as np

from .test_accel import device_test, force_autotune
from .. import accel
from .. import percentile


class TestPercentile5(object):
    def test_percentile(self):
        yield self.check_percentile5, 4096, 1, False, None
        yield self.check_percentile5, 4095, 4029, True, (0, 4009)
        yield self.check_percentile5, 4094, 4030, False, (100, 4030)
        yield self.check_percentile5, 2343, 6031, False, (123, 4001)
        yield self.check_percentile5, 4092, 4032, True, None

    @classmethod
    def pad_dimension(cls, dim, extra):
        """Modifies `dim` to have at least `extra` padding"""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @device_test
    def check_percentile5(self, R, C, is_amplitude, column_range, context, queue):
        template = percentile.Percentile5Template(context, max_columns=5000,
                                                  is_amplitude=is_amplitude)
        fn = template.instantiate(queue, (R, C), column_range)
        # Force some padded, to check that stride calculation works
        self.pad_dimension(fn.slots['src'].dimensions[0], 1)
        self.pad_dimension(fn.slots['src'].dimensions[1], 4)
        self.pad_dimension(fn.slots['dest'].dimensions[0], 2)
        self.pad_dimension(fn.slots['dest'].dimensions[1], 3)
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            ary = np.abs(rs.randn(R, C)).astype(np.float32)  # note positive numbers required
        else:
            ary = (rs.randn(R, C) + 1j * rs.randn(R, C)).astype(np.complex64)
        src = fn.slots['src'].allocate(fn.allocator)
        dest = fn.slots['dest'].allocate(fn.allocator)
        src.set_async(queue, ary)
        fn()
        out = dest.get(queue)
        if column_range is None:
            column_range = (0, C)
        expected = np.percentile(np.abs(ary[:, column_range[0]:column_range[1]]),
                                 [0, 100, 25, 75, 50], axis=1, interpolation='lower')
        expected = expected.astype(dtype=np.float32)
        # When amplitudes are being computed, we won't get a bit-exact match
        if is_amplitude:
            np.testing.assert_equal(expected, out)
        else:
            np.testing.assert_allclose(expected, out, 1e-6)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that the autotuner runs successfully"""
        percentile.Percentile5Template(context, max_columns=5000)



import sys
import functools
import inspect
import shutil
import os
import tempfile
from textwrap import dedent
from unittest import mock

import numpy as np
from decorator import decorator
from mako.template import Template

from nose.tools import assert_equal, assert_raises
from nose.plugins.skip import SkipTest

from .. import accel, tune
from ..accel import HostArray, DeviceArray, SVMArray, LinenoLexer
if accel.have_cuda:
    import pycuda
if accel.have_opencl:
    import pyopencl


_test_context = None
_test_command_queue = None
_test_initialized = False


def _prepare_device_test():
    global _test_initialized, _test_context, _test_command_queue
    if not _test_initialized:
        try:
            _test_context = accel.create_some_context(False)
            _test_command_queue = _test_context.create_command_queue()
            print("Testing on {0} ({1})".format(
                _test_context.device.name, _test_context.device.platform_name),
                file=sys.stderr)
        except RuntimeError:
            pass  # No devices available
        _test_initialized = True

    if not _test_context:
        raise SkipTest('CUDA/OpenCL not found')


def _device_test_sync(test):
    @functools.wraps(test)
    def wrapper(*args, **kwargs):
        _prepare_device_test()
        with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.stub_autotuner):
            args += (_test_context, _test_command_queue)
            # Make the context current (for CUDA contexts). Ideally the test
            # should not depend on this, but PyCUDA leaks memory if objects
            # are deleted without the context current.
            with _test_context:
                return test(*args, **kwargs)
    return wrapper


def _device_test_async(test):
    @functools.wraps(test)
    async def wrapper(*args, **kwargs):
        _prepare_device_test()
        with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.stub_autotuner):
            args += (_test_context, _test_command_queue)
            with _test_context:
                return await test(*args, **kwargs)
    return wrapper


def device_test(test):
    """Decorator that causes a test to be skipped if a compute device is not
    available, and which disables autotuning. If autotuning is desired, use
    :func:`force_autotune` inside (hence, afterwards on the decorator list)
    this one."""
    if inspect.iscoroutinefunction(test):
        return _device_test_async(test)   # noqa: F821
    else:
        return _device_test_sync(test)


def cuda_test(test):
    """Decorator that causes a test to be skipped if the device is not a CUDA
    device. Put this *after* :meth:`device_test`."""
    @functools.wraps(test)
    def wrapper(*args, **kwargs):
        global _test_context
        if not _test_context.device.is_cuda:
            raise SkipTest('Device is not a CUDA device')
        return test(*args, **kwargs)
    return wrapper


@decorator
def force_autotune(test, *args, **kw):
    """Decorator that disables autotuning for a test. Instead, the test
    value specified for each class is returned."""
    with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.force_autotuner):
        return test(*args, **kw)


# Prevent nose from treating it as a test
device_test.__test__ = False
cuda_test.__test__ = False


class TestLinenoLexer(object):
    def test_escape_filename(self):
        assert_equal(
            r'"abc\"def\\ghi"',
            LinenoLexer._escape_filename(r'abc"def\ghi'))

    def test_render(self):
        source = "line 1\nline 2\nline 3"
        out = Template(source, lexer_cls=LinenoLexer).render()
        assert_equal(dedent(
            """\
            #line 1 "<string>"
            line 1
            #line 2 "<string>"
            line 2
            #line 3 "<string>"
            line 3
            """), out)


class TestHostArray(object):
    cls = HostArray

    def allocate(self, shape, dtype, padded_shape):
        return self.cls(shape, dtype, padded_shape)

    def setup(self):
        self.shape = (17, 13)
        self.padded_shape = (32, 16)
        self.constructed = self.allocate(self.shape, np.int32, self.padded_shape)
        self.view = np.zeros(self.padded_shape)[2:4, 3:5].view(self.cls)
        self.sliced = self.constructed[2:4, 3:5]

    def test_safe(self):
        assert self.cls.safe(self.constructed)
        assert not self.cls.safe(self.view)
        assert not self.cls.safe(self.sliced)
        assert not self.cls.safe(np.zeros(self.shape))


class TestDeviceArray(object):
    cls = DeviceArray

    @device_test
    def setup(self, context, queue):
        self.shape = (17, 13)
        self.padded_shape = (32, 16)
        self.strides = (64, 4)
        self.array = self.cls(
            context=context,
            shape=self.shape,
            dtype=np.int32,
            padded_shape=self.padded_shape)

    @device_test
    def test_strides(self, context, queue):
        assert_equal(self.strides, self.array.strides)

    @device_test
    def test_empty_like(self, context, queue):
        ary = self.array.empty_like()
        assert_equal(self.shape, ary.shape)
        assert_equal(self.strides, ary.strides)
        assert HostArray.safe(ary)

    @device_test
    def test_set_get(self, context, queue):
        ary = np.random.randint(0, 100, self.shape).astype(np.int32)
        self.array.set(queue, ary)
        # Read back results, check that it matches
        buf = np.zeros(self.padded_shape, dtype=np.int32)
        if isinstance(self.array, SVMArray):
            buf[:] = self.array.buffer
        elif context.device.is_cuda:
            with context:
                pycuda.driver.memcpy_dtoh(buf, self.array.buffer.gpudata)
        else:
            pyopencl.enqueue_copy(
                queue._pyopencl_command_queue,
                buf, self.array.buffer.data)
        buf = buf[0:self.shape[0], 0:self.shape[1]]
        np.testing.assert_equal(ary, buf)
        # Check that it matches get
        buf = self.array.get(queue)
        np.testing.assert_equal(ary, buf)

    def _test_copy_region(
            self, context, queue,
            src_shape, src_padded_shape,
            dest_shape, dest_padded_shape,
            src_region, dest_region):
        # copy_region test
        dtype = np.int32
        src = self.cls(context, src_shape, dtype, src_padded_shape)
        dest = self.cls(context, dest_shape, dtype, dest_padded_shape)
        h_src = src.empty_like()
        h_src[...] = np.arange(h_src.size).reshape(src_shape)
        src.set(queue, h_src)
        dest.zero(queue)
        src.copy_region(queue, dest, src_region, dest_region)
        queue.finish()   # Needed for the SVMArray test, since the next line is a CPU copy
        h_dest = dest.get(queue)
        expected = np.zeros_like(h_dest)
        expected[dest_region] = h_src[src_region]
        np.testing.assert_array_equal(expected, h_dest)
        # set_region test
        dest.zero(queue)
        dest.set_region(queue, h_src, dest_region, src_region)
        queue.finish()
        h_dest = dest.get(queue)
        np.testing.assert_array_equal(expected, h_dest)
        # get_region test
        h_dest.fill(0)
        src.get_region(queue, h_dest, src_region, dest_region)
        queue.finish()
        np.testing.assert_array_equal(expected, h_dest)

    @device_test
    def test_copy_region_4d(self, context, queue):
        self._test_copy_region(
            context, queue,
            (7, 5, 12, 19), (8, 9, 14, 25),
            (7, 5, 12, 19), (10, 8, 12, 21),
            np.s_[2:4, :, 3:7:2, 10:], np.s_[1:3, 0:5, 4:8:2, 10:])

    @device_test
    def test_copy_region_0d(self, context, queue):
        self._test_copy_region(context, queue, (), (), (), (), (), ())

    @device_test
    def test_copy_region_1d(self, context, queue):
        self._test_copy_region(
            context, queue, (3, 5), (4, 6), (5, 7), (7, 10),
            np.s_[1, 0:3], np.s_[2, 1:4])

    @device_test
    def test_copy_region_2d(self, context, queue):
        self._test_copy_region(
            context, queue, (3, 5), (4, 6), (5, 7), (7, 10),
            np.s_[:1, 0:3], np.s_[-1:, 1:4])

    @device_test
    def test_copy_region_missing_axes(self, context, queue):
        self._test_copy_region(
            context, queue, (3, 5), (4, 6), (5, 5), (7, 10),
            np.s_[:1], np.s_[-1:])

    @device_test
    def test_copy_region_newaxis(self, context, queue):
        self._test_copy_region(
            context, queue, (10,), None, (10, 10), None,
            np.s_[np.newaxis, 4:6], np.s_[:1, 7:9])

    @device_test
    def test_copy_region_negative(self, context, queue):
        self._test_copy_region(
            context, queue, (10, 12), None, (7, 8), None,
            np.s_[-4, -3:-1], np.s_[-7, -8:-4:2])

    @device_test
    def test_copy_region_errors(self, context, queue):
        # Too many axes
        with assert_raises(IndexError):
            self._test_copy_region(
                context, queue, (10,), None, (10,), None,
                np.s_[3, 4], np.s_[5, 6])
        # Out-of-range single index
        with assert_raises(IndexError):
            self._test_copy_region(
                context, queue, (10,), None, (10,), None,
                np.s_[5], np.s_[10])
        # Out-of-range slice
        with assert_raises(IndexError):
            self._test_copy_region(
                context, queue, (10,), None, (10,), None,
                np.s_[10:12], np.s_[8:10])
        # Empty slice
        with assert_raises(IndexError):
            self._test_copy_region(
                context, queue, (10,), None, (10,), None,
                np.s_[2:2], np.s_[3:3])
        # Negative stride
        with assert_raises(IndexError):
            self._test_copy_region(
                context, queue, (10,), None, (10,), None,
                np.s_[3:0:-1], np.s_[4:1:-1])

    @device_test
    def test_zero(self, context, queue):
        ary = np.random.randint(0x12345678, 0x23456789, self.shape).astype(np.int32)
        self.array.set(queue, ary)
        before = self.array.get(queue)
        self.array.zero(queue)
        after = self.array.get(queue)
        np.testing.assert_equal(ary, before)
        assert_equal(0, np.max(after))

    def _allocate_raw(self, context, n_bytes):
        return context.allocate_raw(n_bytes)

    @device_test
    def test_raw(self, context, queue):
        raw = self._allocate_raw(context, 2048)
        ary = self.cls(
            context=context,
            shape=self.shape,
            dtype=np.int32,
            padded_shape=self.padded_shape,
            raw=raw)
        actual_raw = None
        try:
            # CUDA
            if isinstance(ary, SVMArray):
                actual_raw = ary.base.base
                raw = raw._wrapped
            else:
                actual_raw = ary.buffer.gpudata
        except AttributeError:
            # OpenCL
            actual_raw = ary.buffer.data
        assert actual_raw is raw


class TestPinnedAMD(TestDeviceArray):
    """Run DeviceArray tests forcing `_PinnedAMD` class for pinned memory."""
    @device_test
    def setup(self, context, device):
        context._force_pinned_amd = True
        super(TestPinnedAMD, self).setup()

    @device_test
    def teardown(self, context, device):
        context._force_pinned_amd = False


class TestSVMArrayHost(TestHostArray):
    """Tests SVMArray using the HostArray tests"""
    cls = SVMArray

    def allocate(self, shape, dtype, padded_shape):
        return self.cls(self.context, shape, dtype, padded_shape)

    @device_test
    @cuda_test
    def setup(self, context, queue):
        self.context = context
        super(TestSVMArrayHost, self).setup()


class TestSVMArray(TestDeviceArray):
    """Tests SVMArray using the DeviceArray tests, plus some new ones"""
    cls = SVMArray

    def _allocate_raw(self, context, n_bytes):
        return context.allocate_svm_raw(n_bytes)

    @device_test
    @cuda_test
    def setup(self, context, queue):
        super(TestSVMArray, self).setup()

    @device_test
    @cuda_test
    def test_coherence(self, context, queue):
        """Check that runtime correct transfers data between the host and
        device views."""
        source = """\
            <%include file="/port.mako"/>
            KERNEL void triple(unsigned int *data)
            {
                data[get_global_id(0)] *= 3;
            }"""
        program = accel.build(context, None, source=source)
        kernel = program.get_kernel("triple")
        ary = SVMArray(context, (123,), np.uint32, (128,))
        ary[:] = np.arange(123, dtype=np.uint32)
        queue.enqueue_kernel(kernel, [ary.buffer], (128,), (64,))
        queue.finish()
        np.testing.assert_equal(np.arange(369, step=3, dtype=np.uint32), ary)


class TestSVMAllocator(object):
    @device_test
    @cuda_test
    def test_allocate(self, context, queue):
        allocator = accel.SVMAllocator(context)
        ary = allocator.allocate((12, 34), np.int32, (13, 35))
        assert isinstance(ary, accel.SVMArray)
        assert_equal((12, 34), ary.shape)
        assert_equal(np.int32, ary.dtype)
        assert_equal((13, 35), ary.padded_shape)
        ary.fill(1)  # Just to make sure it doesn't crash

    @device_test
    @cuda_test
    def test_allocate_raw(self, context, queue):
        allocator = accel.SVMAllocator(context)
        raw = allocator.allocate_raw(13 * 35 * 4)
        ary = allocator.allocate((12, 34), np.int32, (13, 35), raw=raw)
        assert isinstance(ary, accel.SVMArray)
        assert_equal((12, 34), ary.shape)
        assert_equal(np.int32, ary.dtype)
        assert_equal((13, 35), ary.padded_shape)
        ary.fill(1)  # Just to make sure it doesn't crash


class TestDimension(object):
    """Tests for :class:`katsdpsigproc.accel.Dimension`"""
    def test_is_power2(self):
        assert accel.Dimension._is_power2(1)
        assert accel.Dimension._is_power2(2)
        assert accel.Dimension._is_power2(32)
        assert not accel.Dimension._is_power2(-1)
        assert not accel.Dimension._is_power2(0)
        assert not accel.Dimension._is_power2(3)
        assert not accel.Dimension._is_power2(5)

    def test_min_padded_round(self):
        """Constructor computes min padded size correctly"""
        dim = accel.Dimension(17, min_padded_round=4)
        assert_equal(20, dim.min_padded_size)
        dim = accel.Dimension(20, min_padded_round=5)
        assert_equal(20, dim.min_padded_size)

    def test_add_align_dtype(self):
        dim = accel.Dimension(20, alignment=8)
        assert_equal(8, dim.alignment)
        dim.add_align_dtype(np.complex64)
        assert_equal(accel.Dimension.ALIGN_BYTES // 8, dim.alignment_hint)
        dim.add_align_dtype(np.uint8)
        assert_equal(accel.Dimension.ALIGN_BYTES, dim.alignment_hint)
        dim.add_align_dtype(np.float32)
        assert_equal(accel.Dimension.ALIGN_BYTES, dim.alignment_hint)

    def test_valid(self):
        """Test `valid` method on non-exact dimension"""
        dim = accel.Dimension(17, min_padded_round=8, alignment=4)
        assert dim.valid(24)
        assert not dim.valid(20)
        assert dim.valid(28)
        assert not dim.valid(30)

    def test_valid_exact(self):
        """Test `valid` method on exact dimension"""
        dim = accel.Dimension(20, alignment=4, exact=True)
        assert dim.valid(20)
        assert not dim.valid(24)
        dim = accel.Dimension(20, min_padded_size=23, exact=True)
        assert dim.valid(23)
        assert not dim.valid(24)
        assert not dim.valid(20)

    @classmethod
    def assert_dimensions_equal(cls, dim1, dim2):
        assert_equal(dim1.size, dim2.size)
        assert_equal(dim1.min_padded_size, dim2.min_padded_size)
        assert_equal(dim1.alignment, dim2.alignment)
        assert_equal(dim1.alignment_hint, dim2.alignment_hint)
        assert_equal(dim1.exact, dim2.exact)

    def test_link(self):
        """Linking several dimensions together gives them all the right
        properties."""
        dim1 = accel.Dimension(22, min_padded_size=28, alignment=4)
        dim2 = accel.Dimension(22, min_padded_size=24, alignment=8, align_dtype=np.int32)
        dim3 = accel.Dimension(22, min_padded_size=22, align_dtype=np.uint16)
        dim1.link(dim2)
        dim1.link(dim3)
        self.assert_dimensions_equal(dim1, dim2)
        self.assert_dimensions_equal(dim2, dim3)
        assert_equal(22, dim1.size)
        assert_equal(28, dim1.min_padded_size)
        assert_equal(8, dim1.alignment)
        assert_equal(accel.Dimension.ALIGN_BYTES // 2, dim1.alignment_hint)
        assert_equal(False, dim1.exact)

    def test_link_bad_size(self):
        """Linking dimensions with different sizes fails"""
        dim1 = accel.Dimension(22, min_padded_size=28, alignment=4)
        dim2 = accel.Dimension(23, min_padded_size=24, alignment=8, align_dtype=np.int32)
        with assert_raises(ValueError):
            dim1.link(dim2)
        assert dim1._root() is not dim2._root()

    def test_link_bad_exact(self):
        """Linking dimensions into an unsatisfiable requirement fails"""
        dim1 = accel.Dimension(22, exact=True)
        dim2 = accel.Dimension(22, min_padded_size=28)
        dim3 = accel.Dimension(22, alignment=4)
        with assert_raises(ValueError):
            dim1.link(dim2)
        with assert_raises(ValueError):
            dim1.link(dim3)
        # Check that linking didn't happen anyway
        assert dim1._root() is not dim2._root()
        assert dim1._root() is not dim3._root()

    def test_required_padded_size(self):
        """The padded size is computed correctly"""
        dim = accel.Dimension(30, 7, alignment=4)
        assert_equal(36, dim.required_padded_size())

    def test_required_padded_size_dtype(self):
        """The padded size is computed correctly when an alignment hint is given"""
        dim = accel.Dimension(1100, 200, align_dtype=np.float32)
        assert_equal(1216, dim.required_padded_size())

    def test_required_padded_size_exact(self):
        """The padded size is computed correctly for exact dimensions"""
        dim = accel.Dimension(1100, align_dtype=np.float32, exact=True)
        assert_equal(1100, dim.required_padded_size())

    def test_required_padded_size_small(self):
        """The alignment hint is ignored for small sizes"""
        dim = accel.Dimension(18, alignment=8, align_dtype=np.uint8)
        assert_equal(24, dim.required_padded_size())


class TestIOSlot(object):
    """Tests for :class:`katsdpsigproc.accel.IOSlot`"""
    @mock.patch('katsdpsigproc.accel.DeviceArray', spec=True)
    def test_allocate(self, DeviceArray):
        """IOSlot.allocate must correctly allocate a buffer"""
        dims = [
            accel.Dimension(50, min_padded_size=60, alignment=8),
            accel.Dimension(30, min_padded_size=50, alignment=4)
        ]
        shape = (50, 30)
        padded_shape = (64, 52)
        dtype = np.dtype(np.uint8)
        # Create the device array that will be created. We need to populate it
        # with some attributes to allow validation to pass
        ary = mock.Mock()
        ary.dtype = dtype
        ary.shape = shape
        ary.padded_shape = padded_shape
        # Set the mocked DeviceArray class to return this array
        DeviceArray.return_value = ary
        # Run the system under test
        slot = accel.IOSlot(dims, dtype)
        ret = slot.allocate(accel.DeviceAllocator(mock.sentinel.context))
        # Validation
        assert_equal(ary, ret)
        assert_equal(ary, slot.buffer)
        DeviceArray.assert_called_once_with(
            mock.sentinel.context, shape, dtype, padded_shape, None)
        # Check that the inner dimension had a type hint set
        assert dims[1].alignment_hint == accel.Dimension.ALIGN_BYTES

    @mock.patch('katsdpsigproc.accel.DeviceArray', spec=True)
    def test_allocate_raw(self, DeviceArray):
        """Test IOSlot.allocate with a raw parameter"""
        shape = (50, 30)
        dtype = np.dtype(np.float32)
        raw = mock.sentinel.raw
        # Create the device array that will be created. We need to populate it
        # with some attributes to allow validation to pass
        ary = mock.Mock()
        ary.dtype = dtype
        ary.shape = shape
        ary.padded_shape = shape
        ary.raw = raw
        # Set the mocked DeviceArray class to return this array
        DeviceArray.return_value = ary
        # Run the system under test
        slot = accel.IOSlot(shape, dtype)
        slot.allocate(accel.DeviceAllocator(mock.sentinel.context), raw)
        # Validation
        assert_equal(ary, slot.buffer)
        DeviceArray.assert_called_once_with(
            mock.sentinel.context, shape, dtype, shape, raw)

    def test_validate_shape(self):
        """IOSlot.validate must check that the shape matches"""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = (5, 3)
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)  # Should pass

        with assert_raises(ValueError):
            # Wrong shape
            ary.shape = (5, 4)
            slot.bind(ary)
        with assert_raises(ValueError):
            # Wrong dimensions
            ary.shape = (5,)
            slot.bind(ary)

    def test_validate_dtype(self):
        """IOSlot.validate must check that the dtype matches"""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = ary.shape
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)  # Should pass

        ary.dtype = np.dtype(np.float32)  # Equivalent dtype
        slot.bind(ary)  # Should pass

        with assert_raises(TypeError):
            ary.dtype = np.int32
            slot.bind(ary)

    def test_validate_padded_shape(self):
        """IOSlot.validate must check that the padded shape is valid"""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = (8, 6)
        slot = accel.IOSlot((accel.Dimension(5, 8), accel.Dimension(3, 6)), np.float32)
        slot.bind(ary)  # Should pass

        ary.padded_shape = (10, 11)
        with assert_raises(ValueError):
            # Bigger than needed, but not an exact match - must fail
            slot.bind(ary)

    def test_required_bytes(self):
        slot = accel.IOSlot(
            (accel.Dimension(27, alignment=4), accel.Dimension(33, alignment=32)),
            np.float32)
        assert_equal(4 * 28 * 64, slot.required_bytes())

    def test_bind_none(self):
        """IOSlot.bind must accept `None`"""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = ary.shape
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)
        assert slot.buffer is ary
        slot.bind(None)
        assert slot.buffer is None


class TestCompoundIOSlot(object):
    """Tests for :class:`katsdpsigproc.accel.CompoundIOSlot`"""

    def setup(self):
        self.dims1 = [
            accel.Dimension(13, min_padded_size=17, alignment=1),
            accel.Dimension(7, min_padded_size=8, alignment=8),
            accel.Dimension(22, min_padded_size=25, alignment=4)
        ]
        self.dims2 = [
            accel.Dimension(13, min_padded_size=14, alignment=4),
            accel.Dimension(7, min_padded_size=10, alignment=4),
            accel.Dimension(22, min_padded_size=22, alignment=1)
        ]
        self.slot1 = accel.IOSlot(self.dims1, np.float32)
        self.slot2 = accel.IOSlot(self.dims2, np.float32)

    def test_check_empty(self):
        """CompoundIOSlot constructor must reject empty list"""
        with assert_raises(ValueError):
            accel.CompoundIOSlot([])

    def test_validate_shape(self):
        """CompoundIOSlot must check that children have consistent shapes"""
        slot1 = accel.IOSlot((5, 3), np.float32)
        slot2 = accel.IOSlot((5, 4), np.float32)
        with assert_raises(ValueError):
            accel.CompoundIOSlot([slot1, slot2])

    def test_validate_dtype(self):
        """CompoundIOSlot must check that children have consistent data types"""
        slot1 = accel.IOSlot((5, 3), np.float32)
        slot2 = accel.IOSlot((5, 3), np.int32)
        with assert_raises(TypeError):
            accel.CompoundIOSlot([slot1, slot2])

    def test_attributes(self):
        """CompoundIOSlot must correctly combine attributes"""
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        assert_equal((13, 7, 22), slot.shape)
        assert_equal(np.float32, slot.dtype)
        assert_equal(17, self.dims1[0].min_padded_size)
        assert_equal(10, self.dims1[1].min_padded_size)
        assert_equal(25, self.dims1[2].min_padded_size)
        assert_equal(4, self.dims1[0].alignment)
        assert_equal(8, self.dims1[1].alignment)
        assert_equal(4, self.dims1[2].alignment)
        for x, y in zip(self.dims1, self.dims2):
            TestDimension.assert_dimensions_equal(x, y)

    def test_bind(self):
        """CompoundIOSlot.bind must bind children"""
        ary = mock.sentinel.ary
        ary.shape = (13, 7, 22)
        ary.dtype = np.float32
        ary.padded_shape = (20, 16, 28)
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        slot.bind(ary)
        assert_equal(ary, slot.buffer)
        assert_equal(ary, self.slot1.buffer)
        assert_equal(ary, self.slot2.buffer)

    def test_bind_fail(self):
        """CompoundIOSlot.bind must not have side effects if the array is not
        valid for all children.
        """
        ary = mock.sentinel.ary
        ary.shape = (13, 7, 22)
        ary.dtype = np.float32
        ary.padded_shape = (17, 8, 28)
        # Check that the array has the properties we want for the test
        self.slot1.validate(ary)
        with assert_raises(ValueError):
            self.slot2.validate(ary)
        # The actual test
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        with assert_raises(ValueError):
            slot.bind(ary)
        assert slot.buffer is None
        assert self.slot1.buffer is None
        assert self.slot2.buffer is None


class TestAliasIOSlot(object):
    """Tests for :class:`katsdpsigproc.accel.AliasIOSlot`"""

    def setup(self):
        self.slot1 = accel.IOSlot((3, 7), np.float32)
        self.slot2 = accel.IOSlot((5, 3), np.complex64)

    def test_required_bytes(self):
        slot = accel.AliasIOSlot([self.slot1, self.slot2])
        assert_equal(120, slot.required_bytes())

    def test_allocate(self):
        # Set up mocks
        raw = mock.sentinel.raw
        context = mock.NonCallableMock()
        context.allocate_raw = mock.Mock(return_value=raw)
        # Run the test
        slot = accel.AliasIOSlot([self.slot1, self.slot2])
        ret = slot.allocate(accel.DeviceAllocator(context))
        # Validation
        context.allocate_raw.assert_called_once_with(120)
        assert ret is raw
        assert slot.raw is raw
        assert self.slot1.buffer is not None
        assert self.slot2.buffer is not None


class TestVisualizeOperation(object):
    """Tests for :class:`katsdpsigproc.accel.visualize_operation`.

    This is just a basic smoke test to ensure that it runs without crashing.
    A real test requires a human to sanity-check the visual output.
    """
    class Inner(accel.Operation):
        pass

    def setup(self):
        self.tmpdir = tempfile.mkdtemp()

    def teardown(self):
        shutil.rmtree(self.tmpdir)

    @device_test
    def test(self, context, queue):
        inner = accel.Operation(queue)
        inner.slots['foo'] = accel.IOSlot((3, 4), np.float32)
        inner.slots['bar'] = accel.IOSlot((4, 3), np.float32)

        mid = accel.OperationSequence(
            queue,
            [('inner', inner)],
            {'inner_foo': ['inner:foo']})

        outer = accel.OperationSequence(
            queue,
            [('mid', mid)],
            {'mid_foo': ['mid:inner_foo']},
            {'scratch': ['mid:inner:bar']})

        accel.visualize_operation(outer, os.path.join(self.tmpdir, 'test_visualize.gv'))



#!/usr/bin/env python

"""Test script that runs RFI flagging on random or real data."""

from __future__ import division, print_function, absolute_import
import argparse
import time
import sys
import json
import concurrent.futures

import numpy as np

import katsdpsigproc.rfi.host
import katsdpsigproc.rfi.device
import katsdpsigproc.rfi.twodflag
import katsdpsigproc.accel as accel


def generate_data(times, channels, baselines):
    rs = np.random.RandomState(seed=1)
    shape = (channels, baselines) if times is None else (times, channels, baselines)
    # This is done a row at a time to keep memory usage low
    out = np.empty(shape, np.complex64)
    for i in range(shape[0]):
        real = rs.standard_normal(size=shape[1:]).astype(np.float32)
        imag = rs.standard_normal(size=shape[1:]).astype(np.float32)
        out[i] = real + 1j * imag
    return out


def benchmark1d(args, data):
    if args.width % 2 != 1:
        raise argparse.ArgumentError('Width must be odd')
    if data.shape[0] <= args.width:
        raise argparse.ArgumentError('Channels cannot be less than the filter width')

    context = None
    if not args.host:
        try:
            context = accel.create_some_context(True)
        except RuntimeError:
            print("No devices available. Executing on the CPU.", file=sys.stderr)

    if context is None:
        background = katsdpsigproc.rfi.host.BackgroundMedianFilterHost(args.width)
        noise_est = katsdpsigproc.rfi.host.NoiseEstMADHost()
        threshold = katsdpsigproc.rfi.host.ThresholdSumHost(args.sigmas)
        flagger = katsdpsigproc.rfi.host.FlaggerHost(background, noise_est, threshold)
        start = time.time()
        flags = flagger(data)
        end = time.time()
        print("CPU time (ms):", (end - start) * 1000.0)
    else:
        command_queue = context.create_command_queue(profile=True)
        background = katsdpsigproc.rfi.device.BackgroundMedianFilterDeviceTemplate(
            context, args.width)
        noise_est = katsdpsigproc.rfi.device.NoiseEstMADTDeviceTemplate(
            context, 10240)
        threshold = katsdpsigproc.rfi.device.ThresholdSumDeviceTemplate(context)
        template = katsdpsigproc.rfi.device.FlaggerDeviceTemplate(background, noise_est, threshold)
        flagger = template.instantiate(command_queue, data.shape[0], data.shape[1],
                                       threshold_args={'n_sigma': args.sigmas})
        flagger.ensure_all_bound()

        data_device = flagger.buffer('vis')
        flags_device = flagger.buffer('flags')

        data_device.set(command_queue, data)
        # Run once for warmup (allocates memory)
        flagger()
        # Run again, timing it
        command_queue.finish()

        start_time = time.time()
        start_event = command_queue.enqueue_marker()
        flagger()
        end_event = command_queue.enqueue_marker()
        command_queue.finish()
        end_time = time.time()
        flags = flags_device.get(command_queue)
        print("Host time (ms):  ", (end_time - start_time) * 1000.0)
        try:
            device_time = end_event.time_since(start_event) * 1000.0
        except Exception:
            # AMD CPU device doesn't seem to support profiling on marker events
            device_time = 'unknown'
        print("Device time (ms):", device_time)
    return flags


def benchmark2d(args, data):
    flagger = katsdpsigproc.rfi.twodflag.SumThresholdFlagger(**args.params)
    in_flags = np.zeros(data.shape, np.bool_)
    if args.pool == 'none':
        pool = None
    elif args.pool == 'process':
        pool = concurrent.futures.ProcessPoolExecutor(args.workers)
    elif args.pool == 'thread':
        pool = concurrent.futures.ThreadPoolExecutor(args.workers)
    else:
        raise argparse.ArgumentError('unhandled value {} for --pool'.format(args.pool))
    # Warmup
    try:
        flagger.get_flags(data[:2], in_flags[:2], pool=pool)
        start = time.time()
        flags = flagger.get_flags(data, in_flags, pool=pool)
        end = time.time()
        print("CPU time (ms):", (end - start) * 1000.0)
    finally:
        if pool is not None:
            pool.shutdown()
    return flags


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--host', action='store_true',
                        help='Compute on the host instead of with CUDA/OpenCL')

    parser.add_argument_group('Data selection')
    parser.add_argument('--antennas', '-a', type=int, default=7)
    parser.add_argument('--channels', '-c', type=int, default=1024)
    parser.add_argument('--baselines', '-b', type=int, help='(overrides --antennas)')
    parser.add_argument('--times', '-t', type=int, help='time samples (uses 2D flagger)')
    parser.add_argument('--preset', '-p', choices=('small', 'medium', 'big', 'kat7'),
                        help='(overrides other options)')
    parser.add_argument('--file', type=str, help='specify a real data file (.npy)')

    parser.add_argument_group('Parameters (1D flagger)')
    parser.add_argument('--width', '-w', type=int, help='median filter kernel size (must be odd)',
                        default=13)
    parser.add_argument('--sigmas', type=float, help='threshold for detecting RFI', default=11.0)

    parser.add_argument_group('Parameters (2D flagger)')
    parser.add_argument('--params', type=json.loads, default={}, help='JSON dict of parameters')
    parser.add_argument('--pool', choices=('none', 'process', 'thread'),
                        help='parallelization method')
    parser.add_argument('--workers', type=int, help='Number of parallel workers')

    args = parser.parse_args()

    if args.file is not None:
        if args.file.endswith('.npy'):
            data = np.load(args.file)
        else:
            raise argparse.ArgumentError("Don't know how to handle " + args.file)
    else:
        if args.preset is not None:
            args.baselines = None    # Force recomputation from antennas
            if args.preset == 'small':
                args.antennas = 2
                args.channels = max(2 * args.width, 8)
            elif args.preset == 'medium':
                args.antennas = 15
                args.channels = max(2 * args.width, 128)
            elif args.preset == 'kat7':
                args.antennas = 7
                args.channels = 8192
            elif args.preset == 'big':
                args.antennas = 64
                args.channels = 10240
            else:
                raise argparse.ArgumentError('Unexpected value for preset')
        if args.baselines is None:
            # Note: * 2 not / 2 because there are 4 polarisations
            args.baselines = args.antennas * (args.antennas + 1) * 2
        data = generate_data(args.times, args.channels, args.baselines)

    if args.times is None:
        flags = benchmark1d(args, data)
    else:
        flags = benchmark2d(args, data)
    print('{:.4f}% flagged'.format(100.0 * np.sum(flags) / flags.size))


if __name__ == '__main__':
    main()
